""" Agent to extend the number of tasks given the Transformation definition

The following options can be set for the MCExtensionAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN MCExtensionAgent
  :end-before: ##END
  :dedent: 2
  :caption: MCExtensionAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC import S_OK, gLogger
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

__RCSID__ = "$Id$"

AGENT_NAME = "Transformation/MCExtensionAgent"


class MCExtensionAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        self.transClient = TransformationClient()
        agentTSTypes = self.am_getOption("TransformationTypes", [])
        if agentTSTypes:
            self.transformationTypes = sorted(agentTSTypes)
        else:
            self.transformationTypes = sorted(
                Operations().getValue("Transformations/ExtendableTransfTypes", ["MCSimulation", "Simulation"])
            )
        self.maxIterationTasks = self.am_getOption("TasksPerIteration", 50)
        self.maxFailRate = self.am_getOption("MaxFailureRate", 30)
        self.maxWaitingJobs = self.am_getOption("MaxWaitingJobs", 1000)

    #############################################################################
    def initialize(self):
        """Sets defaults"""

        gLogger.info("Will consider the following transformation types: %s" % str(self.transformationTypes))
        gLogger.info("Will create a maximum of %s tasks per iteration" % self.maxIterationTasks)
        gLogger.info(
            "Will not submit tasks for transformations with failure rate greater than %s%%" % (self.maxFailRate)
        )
        gLogger.info("Will not submit tasks for transformations with more than %d waiting jobs" % self.maxWaitingJobs)

        return S_OK()

    #############################################################################
    def execute(self):
        """The MCExtensionAgent execution method."""

        self.enableFlag = self.am_getOption("EnableFlag", "True")
        if not self.enableFlag == "True":
            self.log.info("MCExtensionAgent is disabled by configuration option EnableFlag")
            return S_OK("Disabled via CS flag")

        # Obtain the transformations in Cleaning status and remove any mention of the jobs/files
        res = self.transClient.getTransformations({"Status": "Active", "Type": self.transformationTypes})
        if res["OK"]:
            for transDict in res["Value"]:
                transID = transDict["TransformationID"]
                maxTasks = transDict["MaxNumberOfTasks"]
                self.extendTransformation(transID, maxTasks)
        return S_OK()

    def extendTransformation(self, transID, maxTasks):
        gLogger.info("Considering extension of transformation %d" % transID)
        # Get the current count of tasks submitted for this transformation
        res = self.transClient.getTransformationTaskStats(transID)
        if not res["OK"]:
            if res["Message"] != "No records found":
                gLogger.error("Failed to get task statistics", "%s %s" % (transID, res["Message"]))
                return res
            else:
                statusDict = {}
        else:
            statusDict = res["Value"]
        gLogger.verbose("Current task count for transformation %d" % transID)
        for status in sorted(statusDict):
            statusCount = statusDict[status]
            gLogger.verbose("%s : %s" % (status.ljust(20), str(statusCount).rjust(8)))
        # Determine the number of tasks to be created
        numberOfTasks = self._calculateTaskNumber(maxTasks, statusDict)
        if not numberOfTasks:
            gLogger.info("No tasks required for transformation %d" % transID)
            return S_OK()
        # Extend the transformation by the determined number of tasks
        res = self.transClient.extendTransformation(transID, numberOfTasks)
        if not res["OK"]:
            gLogger.error("Failed to extend transformation", "%s %s" % (transID, res["Message"]))
            return res
        gLogger.info("Successfully extended transformation %d by %d tasks" % (transID, numberOfTasks))
        return S_OK()

    def _calculateTaskNumber(self, maxTasks, statusDict):
        """Utility function"""
        done = statusDict.get("Done", 0)
        failed = statusDict.get("Failed", 0)
        waiting = statusDict.get("Waiting", 0)
        total = statusDict.get("TotalCreated", 0)
        # If the failure rate is higher than acceptable
        if (total != 0) and ((100.0 * float(failed) / float(total)) > self.maxFailRate):
            return 0
        # If we already have enough completed jobs
        if done >= maxTasks:
            return 0
        if waiting > self.maxWaitingJobs:
            return 0
        numberOfTasks = maxTasks - (total - failed)
        if numberOfTasks < 0:
            # this happens when people extend the transformation manually instead of increasing MaxNumberOfTasks
            return 0
        if numberOfTasks > self.maxIterationTasks:
            numberOfTasks = self.maxIterationTasks
        return numberOfTasks



""" Implementation of Module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-wildcard-import,wildcard-import

import copy
import os

# try: # this part to import as part of the DIRAC framework
from DIRAC.Core.Workflow.Parameter import *

__RCSID__ = "$Id$"


class ModuleDefinition(AttributeCollection):
    def __init__(self, type=None, obj=None, parent=None):
        # we can create an object from another module
        # or from the ParameterCollection
        AttributeCollection.__init__(self)
        self.main_class_obj = None  # used for the interpretation only
        # self.module_obj = None     # used for the interpretation only
        self.parent = parent

        if (obj is None) or isinstance(obj, ParameterCollection):
            self.setType("nitgiven")
            self.setDescrShort("")
            self.setDescription("")
            self.setRequired("")
            self.setBody("")
            self.setOrigin("")
            self.setVersion(0.0)
            self.parameters = ParameterCollection(obj)  # creating copy

        elif isinstance(obj, ModuleDefinition):
            self.setType(obj.getType())
            self.setDescrShort(obj.getDescrShort())
            self.setDescription(obj.getDescription())
            self.setBody(obj.getBody())
            self.setRequired(obj.getRequired())
            self.setOrigin(obj.getOrigin())
            self.setVersion(obj.getVersion())
            self.parameters = ParameterCollection(obj.parameters)
        else:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(obj)))
        if type:
            self.setType(type)

    def createCode(self):
        return self.getBody() + "\n"

    def __str__(self):
        return str(type(self)) + ":\n" + AttributeCollection.__str__(self) + self.parameters.__str__()

    def toXML(self):
        ret = "<ModuleDefinition>\n"
        ret = ret + AttributeCollection.toXML(self)
        ret = ret + self.parameters.toXML()
        ret = ret + "</ModuleDefinition>\n"
        return ret

    def toXMLFile(self, outFile):
        if os.path.exists(outFile):
            os.remove(outFile)
        with open(outFile, "w") as xmlfile:
            xmlfile.write(self.toXML())

    def loadCode(self):
        # print 'Loading code of the Module =', self.getType()
        # version 1 - OLD sample
        # ret = compile(self.getBody(),'<string>','exec')
        # eval(ret)
        # return ret #returning ref just in case we might need it
        #
        if self.getBody():  # checking the size of the string
            # version 2 - we assume that each self.body is a module oblect
            # module = new.module(self.getType())    # create empty module object
            # sys.modules[self.getType()] = module   # add reference for the import operator
            # exec self.getBody() in module.__dict__ # execute code itself
            # self.module_obj = module               # save pointer to this module
            # if module.__dict__.has_key(self.getType()):
            #    self.main_class_obj = module.__dict__[self.getType()] # save class object

            # version 3
            # A.T.  Use vars() function to inspect local objects instead of playing with
            #       fake modules. We assume that after the body execution there will be
            #       a class with name "self.getType()" defined in the local scope.
            exec (self.getBody())
            if self.getType() in vars():
                self.main_class_obj = vars()[self.getType()]  # save class object
            else:
                # it is possible to have this class in another module, we have to check for this
                # but it is advisible to use 'from module import class' operator
                # otherwise i could not find the module. But it is possible that
                # in the future I can change this code to do it more wisely
                sErr = (
                    "Can not find class "
                    + self.getType()
                    + " in the module created from the body of the module "
                    + self.getOrigin()
                )
                raise TypeError(sErr)
        else:
            raise TypeError("The body of the Module " + self.getType() + " seems empty")
        return self.main_class_obj


class ModuleInstance(AttributeCollection):
    def __init__(self, name, obj=None, parent=None):
        AttributeCollection.__init__(self)
        self.instance_obj = None  # used for the interpretation only
        self.parent = parent

        if obj is None:
            self.parameters = ParameterCollection()
        elif isinstance(obj, ModuleInstance) or isinstance(obj, ModuleDefinition):
            if name is None:
                self.setName(obj.getName())
            else:
                self.setName(name)
            self.setType(obj.getType())
            self.setDescrShort(obj.getDescrShort())
            self.parameters = ParameterCollection(obj.parameters)
        elif isinstance(obj, ParameterCollection):
            # set attributes
            self.setName(name)
            self.setType("")
            self.setDescrShort("")
            self.parameters = ParameterCollection(obj)
        elif obj is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(obj)))

    def createCode(self, ind=2):
        str = indent(ind) + self.getName() + " = " + self.getType() + "()\n"
        str = str + self.parameters.createParametersCode(ind, self.getName())
        str = str + indent(ind) + self.getName() + ".execute()\n\n"
        return str

    def __str__(self):
        return str(type(self)) + ":\n" + AttributeCollection.__str__(self) + self.parameters.__str__()

    def toXML(self):
        ret = "<ModuleInstance>\n"
        ret = ret + AttributeCollection.toXML(self)
        ret = ret + self.parameters.toXML()
        ret = ret + "</ModuleInstance>\n"
        return ret

    def execute(self, step_parameters, definitions):
        # print 'Executing ModuleInstance ',self.getName(),'of type',self.getType()
        self.instance_obj = definitions[self.getType()].main_class_obj()  # creating instance
        # FIXME: pylint complains that ParameterCollection doesn't have execute. What should this be?
        self.parameters.execute(self.getName())  # pylint: disable=no-member
        self.instance_obj.execute2()


class DefinitionsPool(dict):
    def __init__(self, parent, pool=None):
        dict.__init__(self)
        self.parent = parent  # this is a cache value, we propagate it into next level
        if isinstance(pool, DefinitionsPool):
            for k in pool:
                v = pool[k]
                if isinstance(v, ModuleDefinition):
                    obj = ModuleDefinition(None, v, self.parent)
                elif isinstance(v, StepDefinition):  # pylint: disable=undefined-variable
                    obj = StepDefinition(None, v, self.parent)  # pylint: disable=undefined-variable
                else:
                    raise TypeError(
                        "Error: __init__ Wrong type of object stored in the DefinitionPool " + str(type(pool[v]))
                    )
                self.append(obj)

        elif pool is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(pool)))

    def __setitem__(self, i, obj):
        if i not in self:
            dict.__setitem__(self, i, obj)

    #       print 'We need to write piece of code to replace existent DefinitionsPool.__setitem__()'
    #       print 'For now we ignore it for the', obj.getType()

    def append(self, obj):
        """We add new Definition (Module, Step)"""
        self[obj.getType()] = obj
        obj.setParent(self.parent)
        return obj

    def remove(self, obj):
        del self[obj.getType()]
        obj.setParent(None)

    def compare(self, s):
        if not isinstance(s, DefinitionsPool):
            return False  # chacking types of objects
        if len(s) != len(self):
            return False  # checkin size
        # we need to compare the keys of dictionaries
        if set(self) != set(s):
            return False
        for k in self:
            if (k not in s) or (not self[k].compare(s[k])):
                return False
        return True

    def __str__(self):
        ret = str(type(self)) + ": number of Definitions:" + str(len(self)) + "\n"
        index = 0
        for k in self:
            ret = ret + "definition(" + str(index) + ")=" + str(self[k]) + "\n"
            index = index + 1
        return ret

    def setParent(self, parent):
        self.parent = parent
        # we need to propagate it just in case it was different one
        for k in self:
            self[k].setParent(parent)

    def getParent(self):
        return self.parent

    def updateParents(self, parent):
        self.parent = parent
        for k in self:
            self[k].updateParents(parent)

    def toXML(self):
        ret = ""
        for k in self:
            ret = ret + self[k].toXML()
        return ret

    def createCode(self):
        str = ""
        for k in self:
            # str=str+indent(2)+'# flush code for instance\n'
            str = str + self[k].createCode()
        return str

    def loadCode(self):
        for k in self:
            # load code of the modules
            self[k].loadCode()


class InstancesPool(list):
    def __init__(self, parent, pool=None):
        list.__init__(self)
        self.parent = None  # this is a cache value, we propagate it into next level
        if isinstance(pool, InstancesPool):
            for v in pool:
                # I need to check this fubction
                # if it would be a costructor we coul pass parent into it
                self.append(copy.deepcopy(v))
                if isinstance(v, ModuleInstance):
                    obj = ModuleInstance(None, v, self.parent)
                elif isinstance(v, StepInstance):  # pylint: disable=undefined-variable
                    obj = StepInstance(None, v, self.parent)  # pylint: disable=undefined-variable
                else:
                    raise TypeError(
                        "Error: __init__ Wrong type of object stored in the DefinitionPool " + str(type(pool[v]))
                    )
                self.append(obj)

        elif pool is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(pool)))

    def __str__(self):
        ret = str(type(self)) + ": number of Instances:" + str(len(self)) + "\n"
        index = 0
        for v in self:
            ret = ret + "instance(" + str(index) + ")=" + str(v) + "\n"
            index = index + 1
        return ret

    def setParent(self, parent):
        self.parent = parent
        for v in self:
            v.setParent(parent)

    def getParent(self):
        return self.parent

    def updateParents(self, parent):
        self.parent = parent
        for v in self:
            v.updateParents(parent)

    def append(self, obj):
        list.append(self, obj)
        obj.setParent(self.parent)

    def toXML(self):
        ret = ""
        for v in self:
            ret = ret + v.toXML()
        return ret

    def findIndex(self, name):
        i = 0
        for v in self:
            if v.getName() == name:
                return i
            i = i + 1
        return -1

    def find(self, name):
        for v in self:
            if v.getName() == name:
                return v
        return None

    def delete(self, name):
        for v in self:
            if v.getName() == name:
                self.remove(v)
                v.setParent(None)

    def compare(self, s):
        if not isinstance(s, InstancesPool) or (len(s) != len(self)):
            return False
        for v in self:
            for i in s:
                if v.getName() == i.getName():
                    if not v.compare(i):
                        return False
                    else:
                        break
            else:
                # if we reached this place naturally we can not find matching name
                return False
        return True

    def createCode(self):
        str = ""
        for inst in self:
            str = str + inst.createCode()
            str = str + indent(2) + "# output assignment\n"
            for v in inst.parameters:
                if v.isOutput():
                    str = str + v.createParameterCode(2, "self")
            str = str + "\n"
        return str



"""
  StorageOccupancyPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from DIRAC import S_OK, gLogger
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from DIRAC.AccountingSystem.Client.Types.StorageOccupancy import StorageOccupancy


__RCSID__ = "$Id$"


class StorageOccupancyPlotter(BaseReporter):
    """StorageOccupancyPlotter as extension of BaseReporter."""

    _typeName = "StorageOccupancy"
    _typeKeyFields = [dF[0] for dF in StorageOccupancy().definitionKeyFields]

    # .............................................................................
    # Generic Reporter

    def reporter(self, reportRequest, spaceType, groupingFields=False):
        """reporter.

        Generic method used by all the _report<something> methods. Only one
        method is actually making use of the groupingFields attribute (
        `_reportFreeUsedSpace` ).
        """

        reportRequest["condDict"]["SpaceType"] = spaceType
        if groupingFields:
            reportRequest["condDict"]["grouping"] = ["SpaceType"]
            reportRequest["groupingFields"] = (["%s", "%s"], reportRequest["groupingFields"][1] + ["SpaceType"])

        selectString = self._getSelectStringForGrouping(reportRequest["groupingFields"])
        selectFields = (
            selectString + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Space", "entriesInBucket"],
        )

        if groupingFields:
            # This is a little hack..
            reportRequest["groupingFields"] = ("%s, %s", reportRequest["groupingFields"][1])

        gLogger.warn(reportRequest)

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal

        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)

        accumMaxValue = self._getAccumulationMaxValue(dataDict)
        suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, "bytes")

        # 3rd value, maxValue is not used
        baseDataDict, graphDataDict, __, unitName = suitableUnits

        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    # .............................................................................
    # Generic Plotter

    def plotter(self, reportRequest, plotInfo, filename):
        """plotter.

        Generic plotter method used by all _plot<something> methods.
        """

        startEpoch = reportRequest["startTime"]
        endEpoch = reportRequest["endTime"]
        granularity = plotInfo["granularity"]
        dataDict = plotInfo["graphDataDict"]

        metadata = {
            "title": "Space grouped by %s" % reportRequest["grouping"],
            "starttime": startEpoch,
            "endtime": endEpoch,
            "span": granularity,
            "ylabel": plotInfo["unit"],
        }

        dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
        return self._generateStackedLinePlot(filename, dataDict, metadata)

    # .............................................................................

    _plotFreeUsedSpace = plotter
    _reportFreeUsedSpaceName = "Free and Used Space"

    def _reportFreeUsedSpace(self, reportRequest):
        """_reportFreeUsedSpace."""

        return self.reporter(reportRequest, ["Free", "Used"], groupingFields=True)

    # .............................................................................

    _plotTotalSpace = plotter
    _reportTotalSpaceName = "Total Space"

    def _reportTotalSpace(self, reportRequest):
        """_reportTotalSpace."""

        return self.reporter(reportRequest, ["Total"])

    # .............................................................................

    _plotGuaranteedSpace = plotter
    _reportGuaranteedSpaceName = "Guaranteed Space"

    def _reportGuaranteedSpace(self, reportRequest):
        """_reportGuaranteedSpace."""

        return self.reporter(reportRequest, ["Guaranteed"])

    # .............................................................................

    _plotFreeSpace = plotter
    _reportFreeSpaceName = "Free Space"

    def _reportFreeSpace(self, reportRequest):
        """_reportFreeSpace."""

        return self.reporter(reportRequest, ["Free"])

    # .............................................................................

    _plotUsedSpace = plotter
    _reportUsedSpaceName = "Used Space"

    def _reportUsedSpace(self, reportRequest):
        """_reportUsedSpace."""

        return self.reporter(reportRequest, ["Used"])



#!/usr/bin/env python
"""
Create a DIRAC MoveReplica request to be executed by the RMS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

import os
import time
from hashlib import md5

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


def getLFNList(arg):
    """get list of LFNs"""
    lfnList = []
    if os.path.exists(arg):
        lfnList = [line.split()[0] for line in open(arg).read().splitlines()]
    else:
        lfnList = [arg]
    return list(set(lfnList))


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(" sourceSE:   source SE")
    Script.registerArgument(" LFN:        LFN or file containing a List of LFNs")
    Script.registerArgument(["targetSE:   target SEs"])
    Script.parseCommandLine()

    import DIRAC
    from DIRAC import gLogger

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    args = Script.getPositionalArgs()

    sourceSE = args[0]
    lfnList = getLFNList(args[1])
    targetSEs = list(set([se for targetSE in args[2:] for se in targetSE.split(",")]))

    gLogger.info(
        "Will create request with 'MoveReplica' "
        "operation using %s lfns and %s target SEs" % (len(lfnList), len(targetSEs))
    )

    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.RequestManagementSystem.Client.Request import Request
    from DIRAC.RequestManagementSystem.Client.Operation import Operation
    from DIRAC.RequestManagementSystem.Client.File import File
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from DIRAC.Core.Utilities.List import breakListIntoChunks

    lfnChunks = breakListIntoChunks(lfnList, 100)
    multiRequests = len(lfnChunks) > 1

    error = 0
    count = 0
    reqClient = ReqClient()
    fc = FileCatalog()
    for lfnChunk in lfnChunks:
        metaDatas = fc.getFileMetadata(lfnChunk)
        if not metaDatas["OK"]:
            gLogger.error("unable to read metadata for lfns: %s" % metaDatas["Message"])
            error = -1
            continue
        metaDatas = metaDatas["Value"]
        for failedLFN, reason in metaDatas["Failed"].items():
            gLogger.error("skipping %s: %s" % (failedLFN, reason))
        lfnChunk = set(metaDatas["Successful"])

        if not lfnChunk:
            gLogger.error("LFN list is empty!!!")
            error = -1
            continue

        if len(lfnChunk) > Operation.MAX_FILES:
            gLogger.error("too many LFNs, max number of files per operation is %s" % Operation.MAX_FILES)
            error = -1
            continue

        count += 1

        request = Request()
        request.RequestName = "%s_%s" % (
            md5(repr(time.time()).encode()).hexdigest()[:16],
            md5(repr(time.time()).encode()).hexdigest()[:16],
        )

        moveReplica = Operation()
        moveReplica.Type = "MoveReplica"
        moveReplica.SourceSE = sourceSE
        moveReplica.TargetSE = ",".join(targetSEs)

        for lfn in lfnChunk:
            metaDict = metaDatas["Successful"][lfn]
            opFile = File()
            opFile.LFN = lfn
            opFile.Size = metaDict["Size"]

            if "Checksum" in metaDict:
                # # should check checksum type, now assuming Adler32 (metaDict["ChecksumType"] = 'AD'
                opFile.Checksum = metaDict["Checksum"]
                opFile.ChecksumType = "ADLER32"
            moveReplica.addFile(opFile)

        request.addOperation(moveReplica)

        result = reqClient.putRequest(request)
        if not result["OK"]:
            gLogger.error("Failed to submit Request: %s" % (result["Message"]))
            error = -1
            continue

        if not multiRequests:
            gLogger.always("Request %d submitted successfully" % result["Value"])

    if multiRequests:
        gLogger.always("%d requests have been submitted" % (count))
    DIRAC.exit(error)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-parameters
# Author :  Stuart Paterson
########################################################################
"""
Retrieve parameters associated to the given DIRAC job

Example:
  $ dirac-wms-job-parameters 1
  {'CPU(MHz)': '1596.479',
   'CPUNormalizationFactor': '6.8',
   'CacheSize(kB)': '4096KB',
   'GridCEQueue': 'ce.labmc.inf.utfsm.cl:2119/jobmanager-lcgpbs-prod',
   'HostName': 'wn05.labmc',
   'JobPath': 'JobPath,JobSanity,JobScheduling,TaskQueue',
   'JobSanityCheck': 'Job: 1 JDL: OK,InputData: No input LFNs,  Input Sandboxes: 0, OK.',
   'JobWrapperPID': '599',
   'LocalAccount': 'prod006',
   'LocalBatchID': '',
   'LocalJobID': '277821.ce.labmc.inf.utfsm.cl',
   'MatcherServiceTime': '2.27646398544',
   'Memory(kB)': '858540kB',
   'ModelName': 'Intel(R)Xeon(R)CPU5110@1.60GHz',
   'NormCPUTime(s)': '1.02',
   'OK': 'True',
   'OutputSandboxMissingFiles': 'std.err',
   'PayloadPID': '604',
   'PilotAgent': 'EELADIRAC v1r1; DIRAC v5r12',
   'Pilot_Reference': 'https://lb2.eela.ufrj.br:9000/ktM6WWR1GdkOTm98_hwM9Q',
   'ScaledCPUTime': '115.6',
   'TotalCPUTime(s)': '0.15'}
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    for job in parseArguments(args):

        result = dirac.getJobParameters(job, printOutput=True)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" pytest of Core.Utilities.Dictionaries
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest

# sut
from DIRAC.Core.Utilities.Dictionaries import breakDictionaryIntoChunks


@pytest.mark.parametrize(
    "aDict, chunkSize, expectedSizes",
    [
        ({}, 1, []),
        ({"a": "aa", "b": "bb", "c": "cc", "d": "dd"}, 5, [4]),
        ({"a": "aa", "b": "bb", "c": "cc", "d": "dd"}, 2, [2, 2]),
        ({"a": "aa", "b": "bb", "c": "cc", "d": "dd"}, 1, [1, 1, 1, 1]),
        ({"a": "aa", "b": "bb", "c": "cc", "d": "dd"}, 3, [3, 1]),
    ],
)
def test_breakDictionaryIntoChunks_normal(aDict, chunkSize, expectedSizes):
    """breakDictIntoChunks tests"""
    result = list(breakDictionaryIntoChunks(aDict, chunkSize))
    assert list(map(len, result)) == expectedSizes
    mergedResult = {}
    for partialDict in result:
        mergedResult.update(partialDict)
    assert aDict == mergedResult



#!/bin/env python
"""
Create a production to replicate files from some storage elements to others

:since:  May 31, 2018
:author: A. Sailer
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    """reads command line parameters, makes check and creates replication transformation"""
    from DIRAC import gLogger, exit as dexit
    from DIRAC.TransformationSystem.Utilities.ReplicationCLIParameters import Params

    clip = Params()
    clip.registerSwitches(Script)
    Script.parseCommandLine()

    from DIRAC.TransformationSystem.Utilities.ReplicationTransformation import createDataTransformation

    if not clip.checkSettings(Script)["OK"]:
        gLogger.error("ERROR: Missing settings")
        dexit(1)
    for metaValue in clip.metaValues:
        resCreate = createDataTransformation(
            flavour=clip.flavour,
            targetSE=clip.targetSE,
            sourceSE=clip.sourceSE,
            metaKey=clip.metaKey,
            metaValue=metaValue,
            extraData=clip.extraData,
            extraname=clip.extraname,
            groupSize=clip.groupSize,
            tGroup=clip.groupName,
            plugin=clip.plugin,
            enable=clip.enable,
        )
        if not resCreate["OK"]:
            gLogger.error("Failed to create Transformation", resCreate["Message"])
            dexit(1)

    dexit(0)


if __name__ == "__main__":
    main()



"""
    This is a comment
"""


# try: # this part to import as part of the DIRAC framework
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import xml.sax
from xml.sax.handler import ContentHandler

from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC.Core.Workflow.Step import *
from DIRAC.Core.Workflow.Workflow import Workflow

__RCSID__ = "$Id$"


class WorkflowXMLHandler(ContentHandler):
    def __init__(self, new_wf=None):
        """If new_wf defined, it will be used as root of document"""
        # this is an attribute for the object to be created from the XML document
        self.root = new_wf  # the reference on the all document
        self.stack = None  # to keep last object
        self.strings = None  # to accumulate string object (list of strings) used to split long string

    def startDocument(self):
        # reset the process
        # self.root=None
        self.stack = []
        self.strings = []

    def endDocument(self):
        pass

    def startElement(self, name, attrs):
        # print name ,"startElement", "attr=", attrs.getLength(), attrs.getNames()
        self.clearCharacters()  # clear to remove empty or nonprintable characters

        if name == "Workflow":
            if self.root is None:  # if root not defined by constractor
                self.root = Workflow()
            self.stack.append(self.root)

        elif name == "StepDefinition":
            obj = StepDefinition("TemporaryXMLObject_StepDefinition")
            if self.root is None:  # in case we are saving Step only
                self.root = obj
            self.stack.append(obj)

        elif name == "StepInstance":
            obj = StepInstance("TemporaryXMLObject_StepInstance")
            self.stack.append(obj)

        elif name == "ModuleDefinition":
            obj = ModuleDefinition("TemporaryXMLObject_ModuleDefinition")
            if self.root is None:  # in case we are saving Module only
                self.root = obj
            self.stack.append(obj)

        elif name == "ModuleInstance":
            obj = ModuleInstance("TemporaryXMLObject_ModuleInstance")
            self.stack.append(obj)

        elif name == "Parameter":
            obj = Parameter(
                str(attrs["name"]),
                None,
                str(attrs["type"]),
                str(attrs["linked_module"]),
                str(attrs["linked_parameter"]),
                str(attrs["in"]),
                str(attrs["out"]),
                str(attrs["description"]),
            )
            self.stack.append(obj)

        # TEMPORARY CODE
        # 2020-03-27: this is temporary code since 2007-06-26
        elif (
            name == "origin"
            or name == "version"
            or name == "name"
            or name == "type"
            or name == "value"
            or name == "required"
            or name == "descr_short"
            or name == "name"
            or name == "type"
            or name == "description"
            or name == "body"
        ):
            pass
        else:
            print("UNTREATED! startElement name=", name, "attr=", attrs.getLength(), attrs.getNames())
            pass

    def endElement(self, name):
        # attributes
        if name == "origin":
            self.stack[-1].setOrigin(self.getCharacters())
        elif name == "version":
            self.stack[-1].setVersion(self.getCharacters())
        elif name == "name":
            self.stack[-1].setName(self.getCharacters())
        elif name == "type":
            self.stack[-1].setType(self.getCharacters())
        elif name == "required":
            self.stack[-1].setRequired(self.getCharacters())
        elif name == "descr_short":
            self.stack[-1].setDescrShort(self.getCharacters())
        elif name == "description":
            self.stack[-1].setDescription(self.getCharacters())
        elif name == "body":
            self.stack[-1].setBody(self.getCharacters())
        elif name == "value":
            ch = self.getCharacters()
            # to keep compatibility with the old version
            # were """ was not used for the string
            if self.stack[-1].isTypeString():
                self.stack[-1].setValue(ch)
            else:
                self.stack[-1].setValue(eval(ch))

        # objects
        elif name == "Workflow":
            self.stack.pop()
        elif name == "StepDefinition":
            self.root.step_definitions.append(self.stack.pop())
        elif name == "StepInstance":
            self.root.step_instances.append(self.stack.pop())
        elif name == "ModuleDefinition":
            self.root.addModule(self.stack.pop())
        elif name == "ModuleInstance":
            obj = self.stack.pop()
            self.stack[-1].module_instances.append(obj)
        elif name == "Parameter":
            obj = self.stack.pop()
            self.stack[-1].addParameter(obj)
        else:
            print("UNTREATED! endElement", name)

    def getCharacters(self):
        # combine all strings and clear the list
        ret = "".join(self.strings)
        self.clearCharacters()
        return str(ret)

    def clearCharacters(self):
        del self.strings
        self.strings = []

    def characters(self, content):
        self.strings.append(content)



""" ResourceManagementHandler

  Module that allows users to access the ResourceManagementDB remotely.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from DIRAC import gConfig, S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.ResourceStatusSystem.Utilities import Synchronizer
from DIRAC.ResourceStatusSystem.Service.ResourceStatusHandler import loadResourceStatusComponent

__RCSID__ = "$Id$"


class ResourceManagementHandler(RequestHandler):
    """
  The ResourceManagementHandler exposes the DB front-end functions through a
  XML-RPC server, functionalities inherited from :class:`DIRAC.Core.DISET.Reques\
  tHandler.RequestHandler`

  According to the ResourceManagementDB philosophy, only functions of the type:
  - insert
  - select
  - delete
  - addOrModify

  are exposed. If you need anything more complicated, either look for it on the
  :class:`ResourceManagementClient`, or code it yourself. This way the DB and the
  Service are kept clean and tidied.

  To can use this service on this way, but you MUST NOT DO IT. Use it through the
  :class:`ResourceManagementClient`. If offers in the worst case as good perfor\
  mance as the :class:`ResourceManagementHandler`, if not better.

   >>> from DIRAC.Core.DISET.RPCClient import RPCCLient
   >>> server = RPCCLient("ResourceStatus/ResourceManagement")
  """

    def __init__(self, *args, **kwargs):
        super(ResourceManagementHandler, self).__init__(*args, **kwargs)

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """
        Dynamically loads ResourceManagement database plugin module, as advised by the config,
        (assumes that the module name and a class name are the same)

        :param serviceInfoDict: service info dictionary
        :return: standard Dirac return object
        """
        defaultOption, defaultClass = "ResourceManagementDB", "ResourceManagementDB"
        configValue = getServiceOption(serviceInfoDict, defaultOption, defaultClass)
        result = loadResourceStatusComponent(configValue, configValue)

        if not result["OK"]:
            return result

        cls.db = result["Value"]
        syncObject = Synchronizer.Synchronizer()
        gConfig.addListenerToNewVersionEvent(syncObject.sync)

        return S_OK()

    def __logResult(self, methodName, result):
        """
        Method that writes to log error messages
        """

        if not result["OK"]:
            self.log.error("%s : %s" % (methodName, result["Message"]))

    types_insert = [six.string_types, dict]

    def export_insert(self, table, params):
        """
        This method is a bridge to access :class:`ResourceManagementDB` remotely. It
        does not add neither processing nor validation. If you need to know more
        about this method, you must keep reading on the database documentation.

        :param table: string or dict - should contain the table from which querying
        :paract dict params: arguments for the mysql query. Currently it is being used only for column selection.

        For example: meta = {'columns': ['Name']} will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        self.log.info("insert: %s %s" % (table, params))

        # remove unnecessary key generated by locals()
        del params["self"]

        res = self.db.insert(table, params)
        self.__logResult("insert", res)

        return res

    types_select = [[six.string_types, dict], dict]

    def export_select(self, table, params):
        """
        This method is a bridge to access :class:`ResourceManagementDB` remotely.
        It does not add neither processing nor validation. If you need to know more\
        about this method, you must keep reading on the database documentation.

        :param table: string or dict - should contain the table from which querying
        :paract dict params: arguments for the mysql query. Currently it is being used only for column selection.

        For example: meta = {'columns': ['Name']} will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        params = {k: list(set(v)) if isinstance(v, list) else v for k, v in params.items()}
        self.log.info("select: %s %s" % (table, params))

        res = self.db.select(table, params)
        self.__logResult("select", res)

        return res

    types_delete = [[six.string_types, dict], dict]

    def export_delete(self, table, params):
        """
        This method is a bridge to access :class:`ResourceManagementDB` remotely.\
        It does not add neither processing nor validation. If you need to know more \
        about this method, you must keep reading on the database documentation.

        :param table: string or dict - should contain the table from which querying
        :paract dict params: arguments for the mysql query. Currently it is being used only for column selection.

        For example: meta = {'columns': ['Name']} will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        self.log.info("delete: %s %s" % (table, params))

        res = self.db.delete(table, params)
        self.__logResult("delete", res)

        return res

    types_addOrModify = [[six.string_types, dict], dict]

    def export_addOrModify(self, table, params):
        """
        This method is a bridge to access :class:`ResourceManagementDB` remotely. It does
        not add neither processing nor validation. If you need to know more about
        this method, you must keep reading on the database documentation.

        :param table: string or dict - should contain the table from which querying
        :paract dict params: arguments for the mysql query. Currently it is being used only for column selection.

        For example: meta = {'columns': ['Name']} will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        self.log.info("addOrModify: %s %s" % (table, params))

        res = self.db.addOrModify(table, params)
        self.__logResult("addOrModify", res)

        return res



""" test File Plugin
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import mock
import unittest
import tempfile
import os
import shutil
import errno

from DIRAC import S_OK
from DIRAC.Resources.Storage.StorageElement import StorageElementItem


def mock_StorageFactory_getConfigStorageName(storageName, referenceType, seConfigPath=None):
    resolvedName = storageName
    return S_OK(resolvedName)


def mock_StorageFactory_getConfigStorageOptions(storageName, derivedStorageName=None, seConfigPath=None):
    """Get the options associated to the StorageElement as defined in the CS"""
    optionsDict = {
        "BackendType": "local",
        "ReadAccess": "Active",
        "WriteAccess": "Active",
        "AccessProtocols": ["file"],
        "WriteProtocols": ["file"],
    }
    return S_OK(optionsDict)


def mock_StorageFactory_getConfigStorageProtocols(storageName, derivedStorageName=None, seConfigPath=None):
    """Protocol specific information is present as sections in the Storage configuration"""
    protocolDetails = {
        "Section": {
            "Host": "",
            "Path": "/tmp/se",
            "PluginName": "File",
            "Port": "",
            "Protocol": "file",
            "SpaceToken": "",
            "WSUrl": "",
        }
    }

    return S_OK(protocolDetails)


class TestBase(unittest.TestCase):
    """Base test class. Defines all the method to test"""

    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageName",
        side_effect=mock_StorageFactory_getConfigStorageName,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageOptions",
        side_effect=mock_StorageFactory_getConfigStorageOptions,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._getConfigStorageProtocols",
        side_effect=mock_StorageFactory_getConfigStorageProtocols,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def setUp(
        self,
        mk_getConfigStorageName,
        mk_getConfigStorageOptions,
        mk_getConfigStorageProtocols,
        mk_isLocalSE,
        mk_addAccountingOperation,
    ):
        self.se = StorageElementItem("FAKE")
        self.se.vo = "test"

        self.basePath = tempfile.mkdtemp(dir="/tmp")
        # Update the basePath of the plugin
        self.se.storages[0].basePath = self.basePath

        self.srcPath = tempfile.mkdtemp(dir="/tmp")

        self.destPath = tempfile.mkdtemp(dir="/tmp")

        self.existingFile = "/test/file.txt"
        self.existingFileSize = 0

        self.nonExistingFile = "/test/nonExistingFile.txt"
        self.subDir = "/test/subDir"
        self.subFile = os.path.join(self.subDir, "subFile.txt")
        self.subFileSize = 0

        self.FILES = [self.existingFile, self.nonExistingFile, self.subFile]
        self.DIRECTORIES = [self.subDir]
        self.ALL = self.FILES + self.DIRECTORIES

        with open(os.path.join(self.srcPath, self.existingFile.replace("/test/", "")), "w") as f:
            f.write("I put something in the file so that it has a size\n")
        self.existingFileSize = os.path.getsize(os.path.join(self.srcPath, self.existingFile.replace("/test/", "")))

        assert self.existingFileSize

        os.mkdir(os.path.join(self.srcPath, os.path.basename(self.subDir)))

        with open(os.path.join(self.srcPath, self.subFile.replace("/test/", "")), "w") as f:
            f.write("This one should have a size as well\n")
        self.subFileSize = os.path.getsize(os.path.join(self.srcPath, self.subFile.replace("/test/", "")))

        assert self.subFileSize

    def tearDown(self):
        shutil.rmtree(self.basePath)
        shutil.rmtree(self.srcPath)
        shutil.rmtree(self.destPath)
        pass

    def walkAll(self):
        for dirname in [self.basePath, self.destPath]:
            self.walkPath(dirname)

    def walkPath(self, path):
        for root, dirs, files in os.walk(path):
            print(root)
            print("  dirs")
            for d in dirs:
                print("    ", os.path.join(root, d))
            print("  files")
            for f in files:
                print("    ", os.path.join(root, f))

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_01_getURL(self, mk_isLocalSE, mk_addAccounting):
        """Testing getURL"""
        # Testing the getURL
        res = self.se.getURL(self.ALL)
        self.assertTrue(res["OK"], res)
        self.assertTrue(not res["Value"]["Failed"], res["Value"]["Failed"])
        self.assertTrue(len(res["Value"]["Successful"]) == len(self.ALL))
        for lfn, url in res["Value"]["Successful"].items():
            self.assertEqual(url, self.basePath.rstrip("/") + lfn)

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_02_FileTest(self, mk_isLocalSE, mk_addAccounting):
        """Testing createDirectory"""
        # Putting the files

        def localPutFile(fn, size=0):
            """If fn is '/test/fn.txt', it calls
            { '/test/fn.txt' : /tmp/generatedPath/fn.txt}
            """
            transfDic = {fn: os.path.join(self.srcPath, fn.replace("/test/", ""))}
            return self.se.putFile(transfDic, sourceSize=size)

        # wrong size
        res = localPutFile(self.existingFile, size=-1)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.existingFile in res["Value"]["Failed"], res)
        self.assertTrue("not match" in res["Value"]["Failed"][self.existingFile], res)
        self.assertTrue(not os.path.exists(self.basePath + self.existingFile))

        # Correct size
        res = localPutFile(self.existingFile, size=self.existingFileSize)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.existingFile in res["Value"]["Successful"], res)
        self.assertTrue(os.path.exists(self.basePath + self.existingFile))

        # No size
        res = localPutFile(self.existingFile)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.existingFile in res["Value"]["Successful"], res)
        self.assertTrue(os.path.exists(self.basePath + self.existingFile))

        # No existing source file
        res = localPutFile(self.nonExistingFile)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.nonExistingFile in res["Value"]["Failed"], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)

        # sub file
        res = localPutFile(self.subFile)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.subFile in res["Value"]["Successful"], res)
        self.assertTrue(os.path.exists(self.basePath + self.subFile))

        # Directory
        res = localPutFile(self.subDir)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.subDir in res["Value"]["Failed"])
        self.assertTrue(
            os.strerror(errno.EISDIR) in res["Value"]["Failed"][self.subDir] or
            # Python 3.9.7+ improved the Exception that is raised
            "Directory does not exist" in res["Value"]["Failed"][self.subDir],
            res,
        )

        res = self.se.exists(self.FILES)
        self.assertTrue(res["OK"], res)
        self.assertTrue(not res["Value"]["Failed"], res)
        self.assertTrue(res["Value"]["Successful"][self.existingFile], res)
        self.assertTrue(not res["Value"]["Successful"][self.nonExistingFile], res)

        res = self.se.getFileSize(self.ALL)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][self.existingFile], self.existingFileSize)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.EISDIR) in res["Value"]["Failed"][self.subDir], res)

        res = self.se.getFileMetadata(self.ALL)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.existingFile in res["Value"]["Successful"])
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.EISDIR) in res["Value"]["Failed"][self.subDir], res)

        res = self.se.isFile(self.ALL)
        self.assertTrue(res["OK"], res)
        self.assertTrue(res["Value"]["Successful"][self.existingFile], res)
        self.assertTrue(not res["Value"]["Successful"][self.subDir], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)

        res = self.se.getFile(self.ALL, localPath=self.destPath)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][self.existingFile], self.existingFileSize)
        self.assertTrue(os.path.exists(os.path.join(self.destPath, os.path.basename(self.existingFile))))
        self.assertEqual(res["Value"]["Successful"][self.subFile], self.subFileSize)
        self.assertTrue(os.path.exists(os.path.join(self.destPath, os.path.basename(self.subFile))))
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(
            os.strerror(errno.EISDIR) in res["Value"]["Failed"][self.subDir] or
            # Python 3.9.7+ improved the Exception that is raised
            "Directory does not exist" in res["Value"]["Failed"][self.subDir],
            res,
        )

        res = self.se.removeFile(self.ALL)
        self.assertTrue(res["OK"], res)
        self.assertTrue(res["Value"]["Successful"][self.existingFile])
        self.assertTrue(not os.path.exists(self.basePath + self.existingFile))
        self.assertTrue(res["Value"]["Successful"][self.subFile])
        self.assertTrue(not os.path.exists(self.basePath + self.subFile))
        self.assertTrue(res["Value"]["Successful"][self.nonExistingFile])
        self.assertTrue(os.strerror(errno.EISDIR) in res["Value"]["Failed"][self.subDir])

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_03_createDirectory(self, mk_isLocalSE, mk_addAccounting):
        """Testing creating directories"""

        res = self.se.createDirectory(self.subDir)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.subDir in res["Value"]["Successful"])
        self.assertTrue(os.path.exists(self.basePath + self.subDir))

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_04_putDirectory(self, mk_isLocalSE, mk_addAccounting):
        """Testing putDirectory"""

        nonExistingDir = "/test/forsuredoesnotexist"
        localdirs = ["/test", nonExistingDir]

        # Correct size
        res = self.se.putDirectory({"/test": self.srcPath})
        self.assertTrue(res["OK"], res)
        self.assertTrue("/test" in res["Value"]["Successful"], res)
        self.assertEqual(
            res["Value"]["Successful"]["/test"], {"Files": 2, "Size": self.existingFileSize + self.subFileSize}
        )
        self.assertTrue(os.path.exists(self.basePath + "/test"))
        self.assertTrue(os.path.exists(self.basePath + self.existingFile))
        self.assertTrue(os.path.exists(self.basePath + self.subFile))

        # No existing source directory
        res = self.se.putDirectory({"/test": nonExistingDir})
        self.assertTrue(res["OK"], res)
        self.assertTrue("/test" in res["Value"]["Failed"], res)
        self.assertEqual(res["Value"]["Failed"]["/test"], {"Files": 0, "Size": 0})

        # sub file
        res = self.se.putDirectory({"/test": self.existingFile})
        self.assertTrue(res["OK"], res)
        self.assertTrue("/test" in res["Value"]["Failed"], res)
        self.assertEqual(res["Value"]["Failed"]["/test"], {"Files": 0, "Size": 0})

        res = self.se.exists(self.DIRECTORIES + localdirs)
        self.assertTrue(res["OK"], res)
        self.assertTrue(not res["Value"]["Failed"], res)
        self.assertTrue(res["Value"]["Successful"][self.subDir], res)
        self.assertTrue(not res["Value"]["Successful"][nonExistingDir], res)

        res = self.se.getDirectorySize(self.ALL + localdirs)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][self.subDir], {"Files": 1, "Size": self.subFileSize, "SubDirs": 0})
        self.assertEqual(res["Value"]["Successful"]["/test"], {"Files": 1, "Size": self.existingFileSize, "SubDirs": 1})
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.ENOTDIR) in res["Value"]["Failed"][self.existingFile], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][nonExistingDir], res)

        res = self.se.getDirectoryMetadata(self.ALL + localdirs)
        self.assertTrue(res["OK"], res)
        self.assertTrue(self.subDir in res["Value"]["Successful"])
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][nonExistingDir], res)
        self.assertTrue(os.strerror(errno.ENOTDIR) in res["Value"]["Failed"][self.existingFile], res)

        res = self.se.isDirectory(self.ALL + localdirs)
        self.assertTrue(res["OK"], res)
        self.assertTrue(not res["Value"]["Successful"][self.existingFile])
        self.assertTrue(res["Value"]["Successful"][self.subDir], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][nonExistingDir], res)

        res = self.se.listDirectory(self.ALL + localdirs)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][self.subDir], {"Files": [self.subFile], "SubDirs": []})
        self.assertEqual(res["Value"]["Successful"]["/test"], {"Files": [self.existingFile], "SubDirs": [self.subDir]})
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][self.nonExistingFile], res)
        self.assertTrue(os.strerror(errno.ENOTDIR) in res["Value"]["Failed"][self.existingFile], res)
        self.assertTrue(os.strerror(errno.ENOENT) in res["Value"]["Failed"][nonExistingDir], res)

        res = self.se.getDirectory(self.ALL + localdirs, localPath=self.destPath)
        self.assertTrue(res["OK"], res)
        self.assertEqual(
            res["Value"]["Successful"]["/test"], {"Files": 2, "Size": self.existingFileSize + self.subFileSize}
        )
        self.assertTrue(os.path.exists(self.destPath + self.existingFile))
        self.assertTrue(os.path.exists(self.destPath + self.subFile))
        self.assertEqual(res["Value"]["Successful"][self.subDir], {"Files": 1, "Size": self.subFileSize})
        self.assertTrue(os.path.exists(self.destPath + self.subFile.replace("/test", "")))
        self.assertEqual(res["Value"]["Failed"][self.nonExistingFile], {"Files": 0, "Size": 0})
        self.assertEqual(res["Value"]["Failed"][self.existingFile], {"Files": 0, "Size": 0})
        self.assertEqual(res["Value"]["Failed"][nonExistingDir], {"Files": 0, "Size": 0})

        res = self.se.removeDirectory(nonExistingDir, recursive=False)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][nonExistingDir], True)

        res = self.se.removeDirectory(nonExistingDir, recursive=True)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Failed"][nonExistingDir], {"FilesRemoved": 0, "SizeRemoved": 0})

        res = self.se.removeDirectory(self.nonExistingFile, recursive=False)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"][self.nonExistingFile], True)

        res = self.se.removeDirectory(self.nonExistingFile, recursive=True)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Failed"][self.nonExistingFile], {"FilesRemoved": 0, "SizeRemoved": 0})

        res = self.se.removeDirectory(self.existingFile, recursive=False)
        self.assertTrue(res["OK"], res)
        self.assertTrue(os.strerror(errno.ENOTDIR) in res["Value"]["Failed"][self.existingFile], res)

        res = self.se.removeDirectory(self.existingFile, recursive=True)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Failed"][self.existingFile], {"FilesRemoved": 0, "SizeRemoved": 0})

        res = self.se.removeDirectory("/test", recursive=False)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"]["/test"], True)
        self.assertTrue(not os.path.exists(self.basePath + self.existingFile))
        self.assertTrue(os.path.exists(self.basePath + self.subFile))

        res = self.se.removeDirectory("/test", recursive=True)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"]["Successful"]["/test"], {"FilesRemoved": 1, "SizeRemoved": self.subFileSize})
        self.assertTrue(not os.path.exists(self.basePath + "/test"))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestBase)

    unittest.TextTestRunner(verbosity=2).run(suite)



########################################################################
# File :    AgentModule.py
# Author :  Adria Casajus
########################################################################
"""
  Base class for all agent modules
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import threading
import time
import signal

import DIRAC
from DIRAC import S_OK, S_ERROR, gConfig, gLogger, rootPath
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Utilities import Time, MemStat, Network
from DIRAC.Core.Utilities.Shifter import setupShifterProxyInEnv
from DIRAC.Core.Utilities.ReturnValues import isReturnStructure
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.FrameworkSystem.Client.MonitoringClient import MonitoringClient
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


class AgentModule(object):
    """Base class for all agent modules

    This class is used by the AgentReactor Class to steer the execution of
    DIRAC Agents.

    For this purpose the following methods are used:
    - am_initialize()      just after instantiated
    - am_getPollingTime()  to set the execution frequency
    - am_getMaxCycles()    to determine the number of cycles
    - am_go()              for the actual execution of one cycle

    Before each iteration, the following methods are used to determine
    if the new cycle is to be started.
    - am_getModuleParam( 'alive' )
    - am_checkStopAgentFile()
    - am_removeStopAgentFile()

    To start new execution cycle the following methods are used
    - am_getCyclesDone()
    - am_setOption( 'MaxCycles', maxCycles )

    At the same time it provides all Agents with common interface.
    All Agent class must inherit from this base class and must implement
    at least the following method:
    - execute()            main method called in the agent cycle

    Additionally they may provide:
    - initialize()         for initial settings
    - finalize()           the graceful exit

    - beginExecution()     before each execution cycle
    - endExecution()       at the end of each execution cycle

    The agent can be stopped either by a signal or by creating a 'stop_agent' file
    in the controlDirectory defined in the agent configuration

    """

    def __init__(self, agentName, loadName, baseAgentName=False, properties={}):
        """
        Common __init__ method for all Agents.
        All Agent modules must define:
        __doc__
        __RCSID__
        They are used to populate __codeProperties

        The following Options are used from the Configuration:
        - /LocalSite/InstancePath
        - /DIRAC/Setup
        - Status
        - Enabled
        - PollingTime            default = 120
        - MaxCycles              default = 500
        - WatchdogTime           default = 0 (disabled)
        - ControlDirectory       control/SystemName/AgentName
        - WorkDirectory          work/SystemName/AgentName
        - shifterProxy           ''
        - shifterProxyLocation   WorkDirectory/SystemName/AgentName/.shifterCred

        It defines the following default Options that can be set via Configuration (above):
        - MonitoringEnabled     True
        - Enabled               True if Status == Active
        - PollingTime           120
        - MaxCycles             500
        - ControlDirectory      control/SystemName/AgentName
        - WorkDirectory         work/SystemName/AgentName
        - shifterProxy          False
        - shifterProxyLocation  work/SystemName/AgentName/.shifterCred

        different defaults can be set in the initialize() method of the Agent using am_setOption()

        In order to get a shifter proxy in the environment during the execute()
        the configuration Option 'shifterProxy' must be set, a default may be given
        in the initialize() method.
        """
        if baseAgentName and agentName == baseAgentName:
            self.log = gLogger
            standaloneModule = True
        else:
            self.log = gLogger.getSubLogger(agentName, child=False)
            standaloneModule = False

        self.__basePath = gConfig.getValue("/LocalSite/InstancePath", rootPath)
        self.__agentModule = None
        self.__codeProperties = {}
        self.__getCodeInfo()

        self.__moduleProperties = {
            "fullName": agentName,
            "loadName": loadName,
            "section": PathFinder.getAgentSection(agentName),
            "loadSection": PathFinder.getAgentSection(loadName),
            "standalone": standaloneModule,
            "cyclesDone": 0,
            "totalElapsedTime": 0,
            "setup": gConfig.getValue("/DIRAC/Setup", "Unknown"),
            "alive": True,
        }
        self.__moduleProperties["system"], self.__moduleProperties["agentName"] = agentName.split("/")
        self.__configDefaults = {}
        self.__configDefaults["MonitoringEnabled"] = True
        self.__configDefaults["Enabled"] = self.am_getOption("Status", "Active").lower() in ("active")
        self.__configDefaults["PollingTime"] = self.am_getOption("PollingTime", 120)
        self.__configDefaults["MaxCycles"] = self.am_getOption("MaxCycles", 500)
        self.__configDefaults["WatchdogTime"] = self.am_getOption("WatchdogTime", 0)
        self.__configDefaults["ControlDirectory"] = os.path.join(self.__basePath, "control", *agentName.split("/"))
        self.__configDefaults["WorkDirectory"] = os.path.join(self.__basePath, "work", *agentName.split("/"))
        self.__configDefaults["shifterProxy"] = ""
        self.__configDefaults["shifterProxyLocation"] = os.path.join(
            self.__configDefaults["WorkDirectory"], ".shifterCred"
        )

        if isinstance(properties, dict):
            for key in properties:
                self.__moduleProperties[key] = properties[key]
            self.__moduleProperties["executors"] = [(self.execute, ())]
            self.__moduleProperties["shifterProxy"] = False

        self.__monitorLastStatsUpdate = -1
        self.monitor = None
        self.__initializeMonitor()
        self.__initialized = False

    def __getCodeInfo(self):
        versionVar = "__RCSID__"
        docVar = "__doc__"
        try:
            self.__agentModule = __import__(self.__class__.__module__, globals(), locals(), versionVar)
        except Exception as excp:
            self.log.exception("Cannot load agent module", lException=excp)
        for prop in ((versionVar, "version"), (docVar, "description")):
            try:
                self.__codeProperties[prop[1]] = getattr(self.__agentModule, prop[0])
            except Exception:
                self.log.error("Missing property", prop[0])
                self.__codeProperties[prop[1]] = "unset"
        self.__codeProperties["DIRACVersion"] = DIRAC.version
        self.__codeProperties["platform"] = DIRAC.getPlatform()

    def am_initialize(self, *initArgs):
        """Common initialization for all the agents.

        This is executed every time an agent (re)starts.
        This is called by the AgentReactor, should not be overridden.
        """
        agentName = self.am_getModuleParam("fullName")
        result = self.initialize(*initArgs)
        if not isReturnStructure(result):
            return S_ERROR("initialize must return S_OK/S_ERROR")
        if not result["OK"]:
            return S_ERROR("Error while initializing %s: %s" % (agentName, result["Message"]))
        mkDir(self.am_getControlDirectory())
        workDirectory = self.am_getWorkDirectory()
        mkDir(workDirectory)
        # Set the work directory in an environment variable available to subprocesses if needed
        os.environ["AGENT_WORKDIRECTORY"] = workDirectory

        self.__moduleProperties["shifterProxy"] = self.am_getOption("shifterProxy")
        if self.am_monitoringEnabled() and not self.activityMonitoring:
            self.monitor.enable()
        if len(self.__moduleProperties["executors"]) < 1:
            return S_ERROR("At least one executor method has to be defined")
        if not self.am_Enabled():
            return S_ERROR("Agent is disabled via the configuration")
        self.log.notice("=" * 40)
        self.log.notice("Loaded agent module %s" % self.__moduleProperties["fullName"])
        self.log.notice(" Site: %s" % DIRAC.siteName())
        self.log.notice(" Setup: %s" % gConfig.getValue("/DIRAC/Setup"))
        self.log.notice(" Base Module version: %s " % __RCSID__)
        self.log.notice(" Agent version: %s" % self.__codeProperties["version"])
        self.log.notice(" DIRAC version: %s" % DIRAC.version)
        self.log.notice(" DIRAC platform: %s" % DIRAC.getPlatform())
        pollingTime = int(self.am_getOption("PollingTime"))
        if pollingTime > 3600:
            self.log.notice(" Polling time: %s hours" % (pollingTime / 3600.0))
        else:
            self.log.notice(" Polling time: %s seconds" % self.am_getOption("PollingTime"))
        self.log.notice(" Control dir: %s" % self.am_getControlDirectory())
        self.log.notice(" Work dir: %s" % self.am_getWorkDirectory())
        if self.am_getOption("MaxCycles") > 0:
            self.log.notice(" Cycles: %s" % self.am_getMaxCycles())
        else:
            self.log.notice(" Cycles: unlimited")
        if self.am_getWatchdogTime() > 0:
            self.log.notice(" Watchdog interval: %s" % self.am_getWatchdogTime())
        else:
            self.log.notice(" Watchdog interval: disabled ")
        self.log.notice("=" * 40)
        self.__initialized = True
        return S_OK()

    def am_getControlDirectory(self):
        return os.path.join(self.__basePath, str(self.am_getOption("ControlDirectory")))

    def am_getStopAgentFile(self):
        return os.path.join(self.am_getControlDirectory(), "stop_agent")

    def am_checkStopAgentFile(self):
        return os.path.isfile(self.am_getStopAgentFile())

    def am_createStopAgentFile(self):
        try:
            with open(self.am_getStopAgentFile(), "w") as fd:
                fd.write("Dirac site agent Stopped at %s" % Time.toString())
        except Exception:
            pass

    def am_removeStopAgentFile(self):
        try:
            os.unlink(self.am_getStopAgentFile())
        except Exception:
            pass

    def am_getBasePath(self):
        return self.__basePath

    def am_getWorkDirectory(self):
        return os.path.join(self.__basePath, str(self.am_getOption("WorkDirectory")))

    def am_getShifterProxyLocation(self):
        return os.path.join(self.__basePath, str(self.am_getOption("shifterProxyLocation")))

    def am_getOption(self, optionName, defaultValue=None):
        """Gets an option from the agent's configuration section.
        The section will be a subsection of the /Systems section in the CS.
        """
        if defaultValue is None:
            if optionName in self.__configDefaults:
                defaultValue = self.__configDefaults[optionName]
        if optionName and optionName[0] == "/":
            return gConfig.getValue(optionName, defaultValue)
        for section in (self.__moduleProperties["section"], self.__moduleProperties["loadSection"]):
            result = gConfig.getOption("%s/%s" % (section, optionName), defaultValue)
            if result["OK"]:
                return result["Value"]
        return defaultValue

    def am_setOption(self, optionName, value):
        self.__configDefaults[optionName] = value

    def am_getModuleParam(self, optionName):
        return self.__moduleProperties[optionName]

    def am_setModuleParam(self, optionName, value):
        self.__moduleProperties[optionName] = value

    def am_getPollingTime(self):
        return self.am_getOption("PollingTime")

    def am_getMaxCycles(self):
        return self.am_getOption("MaxCycles")

    def am_getWatchdogTime(self):
        return int(self.am_getOption("WatchdogTime"))

    def am_getCyclesDone(self):
        return self.am_getModuleParam("cyclesDone")

    def am_Enabled(self):
        return self.am_getOption("Enabled")

    def am_disableMonitoring(self):
        self.am_setOption("MonitoringEnabled", False)

    def am_monitoringEnabled(self):
        return self.am_getOption("MonitoringEnabled")

    def am_stopExecution(self):
        self.am_setModuleParam("alive", False)

    def __initializeMonitor(self):
        """
        Initialize the system monitoring.
        """
        # This flag is used to activate ES based monitoring
        # if the "EnableActivityMonitoring" flag in "yes" or "true" in the cfg file.
        self.activityMonitoring = (
            Operations().getValue("EnableActivityMonitoring", False) and self.am_monitoringEnabled()
        )
        if self.activityMonitoring:
            # The import needs to be here because of the CS must be initialized before importing
            # this class (see https://github.com/DIRACGrid/DIRAC/issues/4793)
            from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

            self.activityMonitoringReporter = MonitoringReporter(monitoringType="ComponentMonitoring")
            # With the help of this periodic task we commit the data to ES at an interval of 100 seconds.
            gThreadScheduler.addPeriodicTask(100, self.__activityMonitoringReporting)
        else:
            if self.__moduleProperties["standalone"]:
                self.monitor = gMonitor
            else:
                self.monitor = MonitoringClient()
            self.monitor.setComponentType(self.monitor.COMPONENT_AGENT)
            self.monitor.setComponentName(self.__moduleProperties["fullName"])
            self.monitor.initialize()
            self.monitor.registerActivity("CPU", "CPU Usage", "Framework", "CPU,%", self.monitor.OP_MEAN, 600)
            self.monitor.registerActivity("MEM", "Memory Usage", "Framework", "Memory,MB", self.monitor.OP_MEAN, 600)
            # Component monitor
            for field in ("version", "DIRACVersion", "description", "platform"):
                self.monitor.setComponentExtraParam(field, self.__codeProperties[field])
            self.monitor.setComponentExtraParam("startTime", Time.dateTime())
            self.monitor.setComponentExtraParam("cycles", 0)
            self.monitor.disable()
            self.__monitorLastStatsUpdate = time.time()

    def am_secureCall(self, functor, args=(), name=False):
        if not name:
            name = str(functor)
        try:
            result = functor(*args)
            if not isReturnStructure(result):
                raise Exception(
                    "%s method for %s module has to return S_OK/S_ERROR" % (name, self.__moduleProperties["fullName"])
                )
            return result
        except Exception as e:
            self.log.exception("Agent exception while calling method %s" % name, lException=e)
            return S_ERROR("Exception while calling %s method: %s" % (name, str(e)))

    def _setShifterProxy(self):
        if self.__moduleProperties["shifterProxy"]:
            result = setupShifterProxyInEnv(self.__moduleProperties["shifterProxy"], self.am_getShifterProxyLocation())
            if not result["OK"]:
                self.log.error("Failed to set shifter proxy", result["Message"])
                return result
        return S_OK()

    def am_go(self):
        # Set the shifter proxy if required
        result = self._setShifterProxy()
        if not result["OK"]:
            return result
        self.log.notice("-" * 40)
        self.log.notice("Starting cycle for module %s" % self.__moduleProperties["fullName"])
        mD = self.am_getMaxCycles()
        if mD > 0:
            cD = self.__moduleProperties["cyclesDone"]
            self.log.notice("Remaining %s of %s cycles" % (mD - cD, mD))
        self.log.notice("-" * 40)
        # use SIGALARM as a watchdog interrupt if enabled
        watchdogInt = self.am_getWatchdogTime()
        if watchdogInt > 0:
            signal.signal(signal.SIGALRM, signal.SIG_DFL)
            signal.alarm(watchdogInt)
        elapsedTime = time.time()
        cpuStats = self._startReportToMonitoring()
        cycleResult = self.__executeModuleCycle()
        if cpuStats:
            self._endReportToMonitoring(*cpuStats)
        # Increment counters
        self.__moduleProperties["cyclesDone"] += 1
        # Show status
        elapsedTime = time.time() - elapsedTime
        self.__moduleProperties["totalElapsedTime"] += elapsedTime
        self.log.notice("-" * 40)
        self.log.notice("Agent module %s run summary" % self.__moduleProperties["fullName"])
        self.log.notice(" Executed %s times previously" % self.__moduleProperties["cyclesDone"])
        self.log.notice(" Cycle took %.2f seconds" % elapsedTime)
        averageElapsedTime = self.__moduleProperties["totalElapsedTime"] / self.__moduleProperties["cyclesDone"]
        self.log.notice(" Average execution time: %.2f seconds" % (averageElapsedTime))
        elapsedPollingRate = averageElapsedTime * 100 / self.am_getOption("PollingTime")
        self.log.notice(" Polling time: %s seconds" % self.am_getOption("PollingTime"))
        self.log.notice(" Average execution/polling time: %.2f%%" % elapsedPollingRate)
        if cycleResult["OK"]:
            self.log.notice(" Cycle was successful")
            if self.activityMonitoring:
                # Here we record the data about the cycle duration along with some basic details about the
                # component and right now it isn't committed to the ES backend.
                self.activityMonitoringReporter.addRecord(
                    {
                        "timestamp": int(Time.toEpoch()),
                        "host": Network.getFQDN(),
                        "componentType": "agent",
                        "component": "_".join(self.__moduleProperties["fullName"].split("/")),
                        "cycleDuration": elapsedTime,
                        "cycles": 1,
                    }
                )
        else:
            self.log.warn(" Cycle had an error:", cycleResult["Message"])
        self.log.notice("-" * 40)
        # Update number of cycles
        if not self.activityMonitoring:
            self.monitor.setComponentExtraParam("cycles", self.__moduleProperties["cyclesDone"])
        # cycle finished successfully, cancel watchdog
        if watchdogInt > 0:
            signal.alarm(0)
        return cycleResult

    def _startReportToMonitoring(self):
        try:
            if not self.activityMonitoring:
                now = time.time()
                stats = os.times()
                cpuTime = stats[0] + stats[2]
                if now - self.__monitorLastStatsUpdate < 10:
                    return (now, cpuTime)
                # Send CPU consumption mark
                self.__monitorLastStatsUpdate = now
                # Send Memory consumption mark
                membytes = MemStat.VmB("VmRSS:")
                if membytes:
                    mem = membytes / (1024.0 * 1024.0)
                    gMonitor.addMark("MEM", mem)
                return (now, cpuTime)
            else:
                return False
        except Exception:
            return False

    def _endReportToMonitoring(self, initialWallTime, initialCPUTime):
        wallTime = time.time() - initialWallTime
        stats = os.times()
        cpuTime = stats[0] + stats[2] - initialCPUTime
        percentage = 0
        if wallTime:
            percentage = cpuTime / wallTime * 100.0
        if percentage > 0:
            gMonitor.addMark("CPU", percentage)

    def __executeModuleCycle(self):
        # Execute the beginExecution function
        result = self.am_secureCall(self.beginExecution, name="beginExecution")
        if not result["OK"]:
            return result
        # Launch executor functions
        executors = self.__moduleProperties["executors"]
        if len(executors) == 1:
            result = self.am_secureCall(executors[0][0], executors[0][1])
            if not result["OK"]:
                return result
        else:
            exeThreads = [threading.Thread(target=executor[0], args=executor[1]) for executor in executors]
            for thread in exeThreads:
                thread.setDaemon(1)
                thread.start()
            for thread in exeThreads:
                thread.join()
        # Execute the endExecution function
        return self.am_secureCall(self.endExecution, name="endExecution")

    def initialize(self, *args, **kwargs):
        """Agents should override this method for specific initialization.
        Executed at every agent (re)start.
        """
        return S_OK()

    def beginExecution(self):
        return S_OK()

    def endExecution(self):
        return S_OK()

    def finalize(self):
        return S_OK()

    def execute(self):
        return S_ERROR("Execute method has to be overwritten by agent module")

    def __activityMonitoringReporting(self):
        """This method is called by the ThreadScheduler as a periodic task in order to commit the collected data which
        is done by the MonitoringReporter and is send to the 'ComponentMonitoring' type.

        :return: True / False
        """
        result = self.activityMonitoringReporter.commit()
        return result["OK"]



#!/usr/bin/env python
"""
Get information on resources available for the given VO: Computing and Storage.
By default, resources for the VO corresponding to the current user identity are displayed
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():

    from DIRAC import S_OK, gLogger, gConfig, exit as DIRACExit

    ceFlag = False
    seFlag = False
    voName = None

    def setCEFlag(args_):
        global ceFlag
        ceFlag = True

    def setSEFlag(args_):
        global seFlag
        seFlag = True

    def setVOName(args):
        global voName
        voName = args

    Script.registerSwitch("C", "ce", "Get CE info", setCEFlag)
    Script.registerSwitch("S", "se", "Get SE info", setSEFlag)
    Script.registerSwitch("V:", "vo=", "Get resources for the given VO. If not set, taken from the proxy", setVOName)

    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
    from DIRAC.ConfigurationSystem.Client.Helpers import Resources
    from DIRAC.Core.Utilities.PrettyPrint import printTable
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
    from DIRAC.Resources.Storage.StorageElement import StorageElement
    from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
    from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus

    def printCEInfo(voName):

        resultQueues = Resources.getQueues(community=voName)
        if not resultQueues["OK"]:
            gLogger.error("Failed to get CE information")
            DIRACExit(-1)

        fields = ("Site", "CE", "CEType", "Queue", "Status")
        records = []

        # get list of usable sites within this cycle
        resultMask = SiteStatus().getUsableSites()
        if not resultMask["OK"]:
            return resultMask
        siteMaskList = resultMask.get("Value", [])

        rssClient = ResourceStatus()

        for site in resultQueues["Value"]:
            siteStatus = "Active" if site in siteMaskList else "InActive"
            siteNew = True
            for ce in resultQueues["Value"][site]:

                ceStatus = siteStatus
                if rssClient.rssFlag:
                    result = rssClient.getElementStatus(ce, "ComputingElement")
                    if result["OK"]:
                        ceStatus = result["Value"][ce]["all"]

                ceNew = True
                for queue in resultQueues["Value"][site][ce]["Queues"]:
                    pSite = site if siteNew else ""
                    pCE = ""
                    ceType = ""
                    if ceNew:
                        pCE = ce
                        ceType = resultQueues["Value"][site][ce]["CEType"]
                    records.append((pSite, pCE, ceType, queue, ceStatus))
                    ceNew = False
                    siteNew = False

        gLogger.notice(printTable(fields, records, printOut=False, columnSeparator="  "))
        return S_OK()

    def printSEInfo(voName):

        fields = ("SE", "Status", "Protocols", "Aliases")
        records = []

        for se in DMSHelpers(
            voName
        ).getStorageElements():  # this will get the full list of SEs, not only the vo's ones.
            seObject = StorageElement(se)
            if not (seObject.vo and voName in seObject.vo.strip().split(",") or not seObject.voName):
                continue

            result = seObject.status()
            status = []
            for statusType in ["Write", "Read"]:
                if result[statusType]:
                    status.append(statusType)

            if status:
                status = "/".join(status)
            else:
                status = "InActive"

            records.append((se, status, ",".join([seProtocol["Protocol"] for seProtocol in seObject.protocolOptions])))

        gLogger.notice(printTable(fields, records, printOut=False, columnSeparator="  "))
        return S_OK()

    if not voName:
        # Get the current VO
        result = getVOfromProxyGroup()
        if not result["OK"]:
            gLogger.error("No proxy found, please login")
            DIRACExit(-1)
        voName = result["Value"]
    else:
        result = gConfig.getSections("/Registry/VO")
        if not result["OK"]:
            gLogger.error("Failed to contact the CS")
            DIRACExit(-1)
        if voName not in result["Value"]:
            gLogger.error("Invalid VO name")
            DIRACExit(-1)

    if not (ceFlag or seFlag):
        gLogger.error("Resource type is not specified")
        DIRACExit(-1)

    if ceFlag:
        result = printCEInfo(voName)
        if not result["OK"]:
            gLogger.error(result["Message"])
            DIRACExit(-1)
    if seFlag:
        result = printSEInfo(voName)
        if not result["OK"]:
            gLogger.error(result["Message"])
            DIRACExit(-1)

    DIRACExit(0)


if __name__ == "__main__":
    main()



########################################################################
# File :   ExecutorReactor.py
# Author : Adria Casajus
########################################################################
"""
  DIRAC class to execute Executors

  Executors are an active part of DIRAC.

  All DIRAC executors must inherit from the basic class ExecutorModule

  In the most common case, DIRAC Executors are executed using the dirac-executor command.
  dirac-execuot accepts a list positional arguments.

  dirac-executo then:
  - produces a instance of ExecutorReactor

  Executor modules must be placed under the Executor directory of a DIRAC System.
  DIRAC Systems are called XXXSystem where XXX is the [DIRAC System Name], and
  must inherit from the base class ExecutorModule

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import threading
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.DISET.MessageClient import MessageClient
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Base.private.ModuleLoader import ModuleLoader
from DIRAC.Core.Base.ExecutorModule import ExecutorModule

__RCSID__ = "$Id$"


class ExecutorReactor(object):
    class AliveLock(object):
        def __init__(self):
            self.__alive = 0
            self.__cond = threading.Condition(threading.Lock())

        def alive(self):
            self.__cond.acquire()
            self.__alive += 1
            self.__cond.release()

        def dead(self):
            self.__cond.acquire()
            self.__alive -= 1
            self.__cond.notify()
            self.__cond.release()

        def lockUntilAllDead(self):
            self.__cond.acquire()
            while True:
                if self.__alive < 1:
                    break
                self.__cond.wait(1)
            self.__cond.release()

    class MindCluster(object):
        def __init__(self, mindName, aliveLock):
            self.__mindName = mindName
            self.__modules = {}
            self.__maxTasks = 1
            self.__reconnectSleep = 1
            self.__reconnectRetries = 10
            self.__extraArgs = {}
            self.__instances = {}
            self.__instanceLock = threading.Lock()
            self.__aliveLock = aliveLock

        def updateMaxTasks(self, mt):
            self.__maxTasks = max(self.__maxTasks, mt)

        def addModule(self, name, exeClass):
            self.__modules[name] = exeClass
            self.__maxTasks = max(self.__maxTasks, exeClass.ex_getOption("MaxTasks", 0))
            self.__reconnectSleep = max(self.__reconnectSleep, exeClass.ex_getOption("ReconnectSleep", 0))
            self.__reconnectRetries = max(self.__reconnectRetries, exeClass.ex_getOption("ReconnectRetries", 0))
            self.__extraArgs[name] = exeClass.ex_getExtraArguments()

        def connect(self):
            self.__msgClient = MessageClient(self.__mindName)
            self.__msgClient.subscribeToMessage("ProcessTask", self.__processTask)
            self.__msgClient.subscribeToDisconnect(self.__disconnected)
            result = self.__msgClient.connect(
                executorTypes=list(self.__modules), maxTasks=self.__maxTasks, extraArgs=self.__extraArgs
            )
            if result["OK"]:
                self.__aliveLock.alive()
                gLogger.info("Connected to %s" % self.__mindName)
            return result

        def __disconnected(self, msgClient):
            retryCount = 0
            while True:
                gLogger.notice("Trying to reconnect to %s" % self.__mindName)
                result = self.__msgClient.connect(
                    executorTypes=list(self.__modules), maxTasks=self.__maxTasks, extraArgs=self.__extraArgs
                )

                if result["OK"]:
                    if retryCount >= self.__reconnectRetries:
                        self.__aliveLock.alive()
                    gLogger.notice("Reconnected to %s" % self.__mindName)
                    return S_OK()
                retryCount += 1
                if retryCount == self.__reconnectRetries:
                    self.__aliveLock.alive()
                gLogger.info("Connect error failed: %s" % result["Message"])
                gLogger.notice("Failed to reconnect. Sleeping for %d seconds" % self.__reconnectSleep)
                time.sleep(self.__reconnectSleep)

        def __storeInstance(self, modName, modObj):
            self.__instanceLock.acquire()
            try:
                self.__instances[modName].append(modObj)
            finally:
                self.__instanceLock.release()

        def __getInstance(self, moduleName):
            self.__instanceLock.acquire()
            try:
                if moduleName not in self.__instances:
                    self.__instances[moduleName] = []
                try:
                    return S_OK(self.__instances[moduleName].pop(0))
                except IndexError:
                    pass
            finally:
                self.__instanceLock.release()
            try:
                modObj = self.__modules[moduleName]
            except KeyError:
                return S_ERROR("Unknown %s executor")
            modInstance = modObj()
            return S_OK(modInstance)

        def __sendExecutorError(self, eType, taskId, errMsg):
            result = self.__msgClient.createMessage("ExecutorError")
            if not result["OK"]:
                return result
            msgObj = result["Value"]
            msgObj.taskId = taskId
            msgObj.errorMsg = errMsg
            msgObj.eType = eType
            return self.__msgClient.sendMessage(msgObj)

        def __processTask(self, msgObj):
            eType = msgObj.eType
            taskId = msgObj.taskId
            taskStub = msgObj.taskStub

            result = self.__moduleProcess(eType, taskId, taskStub)
            if not result["OK"]:
                return self.__sendExecutorError(eType, taskId, result["Message"])
            msgName, taskStub, extra = result["Value"]

            result = self.__msgClient.createMessage(msgName)
            if not result["OK"]:
                return self.__sendExecutorError(
                    eType, taskId, "Can't generate %s message: %s" % (msgName, result["Message"])
                )
            gLogger.verbose("Task %s: Sending %s" % (str(taskId), msgName))
            msgObj = result["Value"]
            msgObj.taskId = taskId
            msgObj.taskStub = taskStub
            if msgName == "TaskError":
                msgObj.errorMsg = extra
                msgObj.eType = eType
            elif msgName == "TaskFreeze":
                msgObj.freezeTime = extra
            return self.__msgClient.sendMessage(msgObj)

        def __moduleProcess(self, eType, taskId, taskStub, fastTrackLevel=0):
            result = self.__getInstance(eType)
            if not result["OK"]:
                return result
            modInstance = result["Value"]
            try:
                result = modInstance._ex_processTask(taskId, taskStub)
            except Exception as excp:
                gLogger.exception("Error while processing task %s" % taskId, lException=excp)
                return S_ERROR("Error processing task %s: %s" % (taskId, excp))

            self.__storeInstance(eType, modInstance)

            if not result["OK"]:
                return S_OK(("TaskError", taskStub, "Error: %s" % result["Message"]))
            taskStub, freezeTime, fastTrackType = result["Value"]
            if freezeTime:
                return S_OK(("TaskFreeze", taskStub, freezeTime))
            if fastTrackType:
                if fastTrackLevel < 10 and fastTrackType in self.__modules:
                    gLogger.notice("Fast tracking task %s to %s" % (taskId, fastTrackType))
                    return self.__moduleProcess(fastTrackType, taskId, taskStub, fastTrackLevel + 1)
                else:
                    gLogger.notice("Stopping %s fast track. Sending back to the mind" % (taskId))

            return S_OK(("TaskDone", taskStub, True))

    #####
    # Start of ExecutorReactor
    #####

    def __init__(self):
        self.__aliveLock = self.AliveLock()
        self.__executorModules = {}
        self.__codeModules = {}
        self.__minds = {}
        self.__loader = ModuleLoader("Executor", PathFinder.getExecutorSection, ExecutorModule)

    def loadModules(self, modulesList, hideExceptions=False):
        """
        Load all modules required in moduleList
        """
        result = self.__loader.loadModules(modulesList, hideExceptions=hideExceptions)
        if not result["OK"]:
            return result
        self.__executorModules = self.__loader.getModules()
        return S_OK()

    # Go!
    def go(self):
        for name in self.__executorModules:
            exeClass = self.__executorModules[name]["classObj"]
            result = exeClass._ex_initialize(name, self.__executorModules[name]["loadName"])
            if not result["OK"]:
                return result
            mind = exeClass.ex_getMind()
            if mind not in self.__minds:
                self.__minds[mind] = self.MindCluster(mind, self.__aliveLock)
            mc = self.__minds[mind]
            mc.addModule(name, exeClass)
        for mindName in self.__minds:
            gLogger.info("Trying to connect to %s" % mindName)
            result = self.__minds[mindName].connect()
            if not result["OK"]:
                return result
        self.__aliveLock.lockUntilAllDead()
        return S_OK()



""" Module that holds the ReportsClient Client class
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import tempfile

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.Utilities.Plotting.FileCoding import codeRequestInFileId


@createClient("Accounting/ReportGenerator")
class ReportsClient(Client):
    def __init__(self, transferClient=None, **kwargs):
        """c'tor"""
        super(ReportsClient, self).__init__(**kwargs)
        self.setServer("Accounting/ReportGenerator")

        self.transferClient = transferClient

    def __getTransferClient(self):
        if not self.transferClient:
            return TransferClient("Accounting/ReportGenerator")
        else:
            return self.transferClient

    def listReports(self, typeName):
        result = self._getRPC().listReports(typeName)
        if "rpcStub" in result:
            del result["rpcStub"]
        return result

    def getReport(self, typeName, reportName, startTime, endTime, condDict, grouping, extraArgs=None):

        if not isinstance(extraArgs, dict):
            extraArgs = {}
        plotRequest = {
            "typeName": typeName,
            "reportName": reportName,
            "startTime": startTime,
            "endTime": endTime,
            "condDict": condDict,
            "grouping": grouping,
            "extraArgs": extraArgs,
        }
        result = self._getRPC().getReport(plotRequest)
        if "rpcStub" in result:
            del result["rpcStub"]
        return result

    def generatePlot(self, typeName, reportName, startTime, endTime, condDict, grouping, extraArgs=None):
        if not isinstance(extraArgs, dict):
            extraArgs = {}
        plotRequest = {
            "typeName": typeName,
            "reportName": reportName,
            "startTime": startTime,
            "endTime": endTime,
            "condDict": condDict,
            "grouping": grouping,
            "extraArgs": extraArgs,
        }
        result = self._getRPC().generatePlot(plotRequest)
        if "rpcStub" in result:
            del result["rpcStub"]
        return result

    def generateDelayedPlot(
        self, typeName, reportName, startTime, endTime, condDict, grouping, extraArgs=None, compress=True
    ):
        if not isinstance(extraArgs, dict):
            extraArgs = {}
        plotRequest = {
            "typeName": typeName,
            "reportName": reportName,
            "startTime": startTime,
            "endTime": endTime,
            "condDict": condDict,
            "grouping": grouping,
            "extraArgs": extraArgs,
        }
        return codeRequestInFileId(plotRequest, compress)

    def getPlotToMem(self, plotName):
        transferClient = self.__getTransferClient()
        tmpFile = tempfile.TemporaryFile()
        retVal = transferClient.receiveFile(tmpFile, plotName)
        if not retVal["OK"]:
            return retVal
        tmpFile.seek(0)
        data = tmpFile.read()
        tmpFile.close()
        return S_OK(data)

    def getPlotToDirectory(self, plotName, dirDestination):
        transferClient = self.__getTransferClient()
        try:
            destFilename = "%s/%s" % (dirDestination, plotName)
            with open(destFilename, "wb") as destFile:
                retVal = transferClient.receiveFile(destFile, plotName)
                if not retVal["OK"]:
                    return retVal
        except Exception as e:
            return S_ERROR("Can't open file %s for writing: %s" % (destFilename, str(e)))
        return S_OK()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from string import Template
import six


class Activity:
    """
    This class is maintained in order to maintain a basic mapping of the sqlite3 database meaning
    what all tables and fields the database uses and it creates a mapping for them which is further
    used by the MonitoringCatalog class.
    """

    dbFields = [
        "activities.unit",
        "activities.type",
        "activities.description",
        "activities.filename",
        "activities.bucketLength",
        "sources.site",
        "sources.componentType",
        "sources.componentLocation",
        "sources.componentName",
    ]

    dbMapping = {}

    def __init__(self, dataList):
        """
        Init an activity
        """
        self.dataList = dataList
        self.groupList = []
        self.groupLabel = ""
        self.__initMapping()
        self.templateMap = {}
        self.scaleFactor = 1
        self.labelTemplate = ""
        for fieldName in self.dbFields:
            capsFieldName = fieldName.split(".")[1].upper()
            self.templateMap[capsFieldName] = self.__getField(fieldName)

    def __initMapping(self):
        """
        Init static maping
        """
        if not self.dbMapping:
            for index in range(len(self.dbFields)):
                self.dbMapping[self.dbFields[index]] = index

    def setBucketScaleFactor(self, scaleFactor):
        """
        Sets a bucket scale factor.
        """
        self.scaleFactor = scaleFactor
        self.__calculateUnit()

    def __calculateUnit(self):
        self.dataList = list(self.dataList)
        unit = self.dataList[self.dbMapping["activities.unit"]].split("/")[0]
        if self.getType() in ("sum"):
            sF = int(self.getBucketLength() * self.scaleFactor) / 60
            if sF == 1:
                unit = "%s/min" % unit
            else:
                unit = "%s/%s mins" % (unit, sF)
        if self.getType() in ("rate"):
            unit = "%s/seconds" % unit
        self.dataList[self.dbMapping["activities.unit"]] = unit
        self.templateMap["UNIT"] = unit

    def setGroup(self, group):
        """
        Set group to which this activity belongs
        """
        self.groupList = group
        self.groupLabel = "Grouped for"
        for fieldName in self.groupList:
            self.groupLabel += " %s," % fieldName
        self.groupLabel = self.groupLabel[:-1]

    def setLabel(self, labelTemplate):
        """
        Sets an activity label
        """
        self.labelTemplate = labelTemplate

    def __getField(self, name):
        """
        Gets a field value
        """
        return self.dataList[self.dbMapping[name]]

    def getUnit(self):
        """
        Gets an activity's unit.
        """
        return self.__getField("activities.unit")

    def getFile(self):
        """
        Gets an activity's file.
        """
        return self.__getField("activities.filename")

    def getType(self):
        """
        Gets an activity's type.
        """
        return self.__getField("activities.type")

    def getDescription(self):
        """
        Gets an activity's unit
        """
        return self.__getField("activities.description")

    def getBucketLength(self):
        """
        Gets an activity's bucket length.
        """
        return self.__getField("activities.bucketLength")

    def getSite(self):
        """
        Gets an source's site.
        """
        return self.__getField("sources.site")

    def getComponentType(self):
        """
        Gets an source's component type.
        """
        return self.__getField("sources.componentType")

    def getComponentName(self):
        """
        Gets an source's component name.
        """
        return self.__getField("sources.componentName")

    def getComponentLocation(self):
        """
        Gets an source's component location.
        """
        return self.__getField("sources.componentLocation")

    def getGroupLabel(self):
        """
        Gets a group's label.
        """
        return self.groupLabel

    def getLabel(self):
        if isinstance(self.labelTemplate, six.text_type):
            self.labelTemplate = self.labelTemplate.encode("utf-8")
        return Template(self.labelTemplate).safe_substitute(self.templateMap)

    def __str__(self):
        return "[%s][%s][%s]" % (self.getLabel(), self.getGroupLabel(), str(self.dataList))

    def __repr__(self):
        return self.__str__()

    def __lt__(self, act):
        label = self.getLabel()
        try:
            return label < act.getLabel()
        except Exception:
            return label < act



""" Helper for the CS Resources section
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from six.moves.urllib import parse as urlparse
from distutils.version import LooseVersion  # pylint: disable=no-name-in-module,import-error

import six

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers import Registry, Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.Core.Utilities.List import uniqueElements, fromChar
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager


gBaseResourcesSection = "/Resources"


def getSites():
    """Get the list of all the sites defined in the CS"""
    result = gConfig.getSections(cfgPath(gBaseResourcesSection, "Sites"))
    if not result["OK"]:
        return result
    grids = result["Value"]
    sites = []
    for grid in grids:
        result = gConfig.getSections(cfgPath(gBaseResourcesSection, "Sites", grid))
        if not result["OK"]:
            return result
        sites += result["Value"]

    return S_OK(sites)


def getSiteCEMapping():
    """:returns: dict of site: list of CEs"""
    res = getSites()
    if not res["OK"]:
        return res
    sites = res["Value"]
    sitesCEsMapping = {}
    for site in sites:
        res = sitesCEsMapping[site] = gConfig.getSections(
            cfgPath(gBaseResourcesSection, "Sites", site.split(".")[0], site, "CEs"), []
        )
        if not res["OK"]:
            gLogger.error("Wrong configuration of CE for site:", site)
            continue
        sitesCEsMapping[site] = res["Value"]
    return S_OK(sitesCEsMapping)


def getCESiteMapping(ceName=""):
    """Returns a mapping of CE -> site
    It assumes that the ce names are unique (as they should be)

    :param str ceName: optional CE name. If not present, the whole mapping will be returned
    :return: S_OK/S_ERROR structure
    """
    res = getSiteCEMapping()
    if not res["OK"]:
        return res
    sitesCEs = res["Value"]
    ceSiteMapping = {}
    for site in sitesCEs:
        for ce in sitesCEs[site]:
            if ceName:
                if ce != ceName:
                    continue
            ceSiteMapping[ce] = site
    return S_OK(ceSiteMapping)


def getGOCSiteName(diracSiteName):
    """
    Get GOC DB site name, given the DIRAC site name, as it is stored in the CS

    :param str diracSiteName: DIRAC site name (e.g. 'LCG.CERN.ch')
    :returns: S_OK/S_ERROR structure
    """
    gocDBName = gConfig.getValue(
        cfgPath(gBaseResourcesSection, "Sites", diracSiteName.split(".")[0], diracSiteName, "Name")
    )
    if not gocDBName:
        return S_ERROR("No GOC site name for %s in CS (Not a grid site ?)" % diracSiteName)
    return S_OK(gocDBName)


def getGOCSites(diracSites=None):

    if diracSites is None:
        diracSites = getSites()
        if not diracSites["OK"]:
            return diracSites
        diracSites = diracSites["Value"]

    gocSites = []

    for diracSite in diracSites:
        gocSite = getGOCSiteName(diracSite)
        if not gocSite["OK"]:
            continue
        gocSites.append(gocSite["Value"])

    return S_OK(list(set(gocSites)))


def getDIRACSiteName(gocSiteName):
    """
    Get DIRAC site name, given the GOC DB site name, as it stored in the CS

    :params str gocSiteName: GOC DB site name (e.g. 'CERN-PROD')
    :returns: S_OK/S_ERROR structure
    """
    res = getSites()
    if not res["OK"]:
        return res
    sitesList = res["Value"]

    tmpList = [
        (site, gConfig.getValue(cfgPath(gBaseResourcesSection, "Sites", site.split(".")[0], site, "Name")))
        for site in sitesList
    ]

    diracSites = [dirac for (dirac, goc) in tmpList if goc == gocSiteName]

    if diracSites:
        return S_OK(diracSites)

    return S_ERROR("There's no site with GOCDB name = %s in DIRAC CS" % gocSiteName)


def getGOCFTSName(diracFTSName):
    """
    Get GOC DB FTS server URL, given the DIRAC FTS server name, as it stored in the CS

    :param str diracFTSName: DIRAC FTS server name (e.g. 'CERN-FTS3')
    :returns: S_OK/S_ERROR structure
    """

    gocFTSName = gConfig.getValue(cfgPath(gBaseResourcesSection, "FTSEndpoints", "FTS3", diracFTSName))
    if not gocFTSName:
        return S_ERROR("No GOC FTS server name for %s in CS (Not a grid site ?)" % diracFTSName)
    return S_OK(gocFTSName)


def getFTS3Servers(hostOnly=False):
    """get list of FTS3 servers that are in CS

    :param bool hostOnly: flag for stripping down the protocol and ports
    """

    csPath = cfgPath(gBaseResourcesSection, "FTSEndpoints/FTS3")
    # We do it in two times to keep the order
    ftsServerNames = gConfig.getOptions(csPath).get("Value", [])

    ftsServers = []
    for name in ftsServerNames:
        serverPath = gConfig.getValue(cfgPath(csPath, name))
        if hostOnly:
            serverPath = urlparse.urlparse(serverPath).hostname
        ftsServers.append(serverPath)

    return S_OK(ftsServers)


def getFTS3ServerDict():
    """:returns: dict of key = server name and value = server url"""
    return gConfig.getOptionsDict(cfgPath(gBaseResourcesSection, "FTSEndpoints/FTS3"))


def getSiteTier(site):
    """
    Return Tier level of the given Site
    """
    result = getSitePath(site)
    if not result["OK"]:
        return result
    sitePath = result["Value"]
    return S_OK(gConfig.getValue(cfgPath(sitePath, "MoUTierLevel"), 2))


def getSitePath(site):
    """
    Return path to the Site section on CS
    """
    result = getSiteGrid(site)
    if not result["OK"]:
        return result
    grid = result["Value"]
    return S_OK(cfgPath(gBaseResourcesSection, "Sites", grid, site))


def getSiteGrid(site):
    """
    Return Grid component from Site Name
    """
    sitetuple = site.split(".")
    if len(sitetuple) != 3:
        return S_ERROR("Wrong Site Name format")
    return S_OK(sitetuple[0])


def getQueue(site, ce, queue):
    """Get parameters of the specified queue"""
    grid = site.split(".")[0]
    result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/CEs/%s" % (grid, site, ce))
    if not result["OK"]:
        return result
    resultDict = result["Value"]

    # Get queue defaults
    result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/CEs/%s/Queues/%s" % (grid, site, ce, queue))
    if not result["OK"]:
        return result
    resultDict.update(result["Value"])

    # Handle tag lists for the queue
    for tagFieldName in ("Tag", "RequiredTag"):
        tags = []
        ceTags = resultDict.get(tagFieldName)
        if ceTags:
            tags = fromChar(ceTags)
        queueTags = resultDict.get(tagFieldName)
        if queueTags:
            queueTags = fromChar(queueTags)
            tags = list(set(tags + queueTags))
        if tags:
            resultDict[tagFieldName] = tags

    resultDict["Queue"] = queue
    return S_OK(resultDict)


def getQueues(siteList=None, ceList=None, ceTypeList=None, community=None, mode=None):
    """Get CE/queue options according to the specified selection"""

    result = gConfig.getSections("/Resources/Sites")
    if not result["OK"]:
        return result

    resultDict = {}

    grids = result["Value"]
    for grid in grids:
        result = gConfig.getSections("/Resources/Sites/%s" % grid)
        if not result["OK"]:
            continue
        sites = result["Value"]
        for site in sites:
            if siteList and site not in siteList:
                continue
            if community:
                comList = gConfig.getValue("/Resources/Sites/%s/%s/VO" % (grid, site), [])
                if comList and community.lower() not in [cl.lower() for cl in comList]:
                    continue
            siteCEParameters = {}
            result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/CEs" % (grid, site))
            if result["OK"]:
                siteCEParameters = result["Value"]
            result = gConfig.getSections("/Resources/Sites/%s/%s/CEs" % (grid, site))
            if not result["OK"]:
                continue
            ces = result["Value"]
            for ce in ces:
                if mode:
                    ceMode = gConfig.getValue(
                        "/Resources/Sites/%s/%s/CEs/%s/SubmissionMode" % (grid, site, ce), "Direct"
                    )
                    if not ceMode or ceMode.lower() != mode.lower():
                        continue
                if ceTypeList:
                    ceType = gConfig.getValue("/Resources/Sites/%s/%s/CEs/%s/CEType" % (grid, site, ce), "")
                    if not ceType or ceType not in ceTypeList:
                        continue
                if ceList and ce not in ceList:
                    continue
                if community:
                    comList = gConfig.getValue("/Resources/Sites/%s/%s/CEs/%s/VO" % (grid, site, ce), [])
                    if comList and community.lower() not in [cl.lower() for cl in comList]:
                        continue
                ceOptionsDict = dict(siteCEParameters)
                result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/CEs/%s" % (grid, site, ce))
                if not result["OK"]:
                    continue
                ceOptionsDict.update(result["Value"])
                result = gConfig.getSections("/Resources/Sites/%s/%s/CEs/%s/Queues" % (grid, site, ce))
                if not result["OK"]:
                    continue
                queues = result["Value"]
                for queue in queues:
                    if community:
                        comList = gConfig.getValue(
                            "/Resources/Sites/%s/%s/CEs/%s/Queues/%s/VO" % (grid, site, ce, queue), []
                        )
                        if comList and community.lower() not in [cl.lower() for cl in comList]:
                            continue
                    resultDict.setdefault(site, {})
                    resultDict[site].setdefault(ce, ceOptionsDict)
                    resultDict[site][ce].setdefault("Queues", {})
                    result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/CEs/%s/Queues/%s" % (grid, site, ce, queue))
                    if not result["OK"]:
                        continue
                    queueOptionsDict = result["Value"]
                    resultDict[site][ce]["Queues"][queue] = queueOptionsDict

    return S_OK(resultDict)


def getCompatiblePlatforms(originalPlatforms):
    """Get a list of platforms compatible with the given list"""
    if isinstance(originalPlatforms, six.string_types):
        platforms = [originalPlatforms]
    else:
        platforms = list(originalPlatforms)

    platforms = list(platform.replace(" ", "") for platform in platforms)

    result = gConfig.getOptionsDict("/Resources/Computing/OSCompatibility")
    if not (result["OK"] and result["Value"]):
        return S_ERROR("OS compatibility info not found")

    platformsDict = dict((k, v.replace(" ", "").split(",")) for k, v in result["Value"].items())  # can be an iterator
    for k, v in platformsDict.items():  # can be an iterator
        if k not in v:
            v.append(k)

    resultList = list(platforms)
    for p in platforms:
        tmpList = platformsDict.get(p, [])
        for pp in platformsDict:
            if p in platformsDict[pp]:
                tmpList.append(pp)
                tmpList += platformsDict[pp]
        if tmpList:
            resultList += tmpList

    return S_OK(uniqueElements(resultList))


def getDIRACPlatform(OSList):
    """Get standard DIRAC platform(s) compatible with the argument.

    NB: The returned value is a list! ordered, in reverse, using distutils.version.LooseVersion
    In practice the "highest" version (which should be the most "desirable" one is returned first)

    :param list OSList: list of platforms defined by resource providers
    :return: a list of DIRAC platforms that can be specified in job descriptions
    """

    # For backward compatibility allow a single string argument
    osList = OSList
    if isinstance(OSList, six.string_types):
        osList = [OSList]

    result = gConfig.getOptionsDict("/Resources/Computing/OSCompatibility")
    if not (result["OK"] and result["Value"]):
        return S_ERROR("OS compatibility info not found")

    platformsDict = dict((k, v.replace(" ", "").split(",")) for k, v in result["Value"].items())  # can be an iterator
    for k, v in platformsDict.items():  # can be an iterator
        if k not in v:
            v.append(k)

    # making an OS -> platforms dict
    os2PlatformDict = dict()
    for platform, osItems in platformsDict.items():  # can be an iterator
        for osItem in osItems:
            if os2PlatformDict.get(osItem):
                os2PlatformDict[osItem].append(platform)
            else:
                os2PlatformDict[osItem] = [platform]

    platforms = []
    for os in osList:
        if os in os2PlatformDict:
            platforms += os2PlatformDict[os]

    if not platforms:
        return S_ERROR("No compatible DIRAC platform found for %s" % ",".join(OSList))

    platforms.sort(key=LooseVersion, reverse=True)

    return S_OK(platforms)


def getDIRACPlatforms():
    """just returns list of platforms defined in the CS"""
    result = gConfig.getOptionsDict("/Resources/Computing/OSCompatibility")
    if not (result["OK"] and result["Value"]):
        return S_ERROR("OS compatibility info not found")
    return S_OK(list(result["Value"]))


def getCatalogPath(catalogName):
    """Return the configuration path of the description for a a given catalog"""
    return "/Resources/FileCatalogs/%s" % catalogName


def getBackendConfig(backendID):
    """Return a backend configuration for a given backend identifier

    :params backendID: string representing a backend identifier. Ex: stdout, file, f02
    """
    return gConfig.getOptionsDict("Resources/LogBackends/%s" % backendID)


def getFilterConfig(filterID):
    """Return a filter configuration for a given filter identifier.

    :params filterID: string representing a filter identifier.
    """
    return gConfig.getOptionsDict("Resources/LogFilters/%s" % filterID)


def getInfoAboutProviders(of=None, providerName=None, option="", section=""):
    """Get the information about providers

    :param str of: provider of what(Id, Proxy or etc.) need to look,
           None, "all" to get list of instance of what this providers
    :param str providerName: provider name,
           None, "all" to get list of providers names
    :param str option: option name that need to get,
           None, "all" to get all options in a section
    :param str section: section path in root section of provider,
           "all" to get options in all sections

    :return: S_OK()/S_ERROR()
    """
    if not of or of == "all":
        result = gConfig.getSections(gBaseResourcesSection)
        if not result["OK"]:
            return result
        return S_OK([i.replace("Providers", "") for i in result["Value"]])
    if not providerName or providerName == "all":
        return gConfig.getSections("%s/%sProviders" % (gBaseResourcesSection, of))
    if not option or option == "all":
        if not section:
            return gConfig.getOptionsDict("%s/%sProviders/%s" % (gBaseResourcesSection, of, providerName))
        elif section == "all":
            resDict = {}
            relPath = "%s/%sProviders/%s/" % (gBaseResourcesSection, of, providerName)
            result = gConfig.getConfigurationTree(relPath)
            if not result["OK"]:
                return result
            for key, value in result["Value"].items():  # can be an iterator
                if value:
                    resDict[key.replace(relPath, "")] = value
            return S_OK(resDict)
        else:
            return gConfig.getSections("%s/%sProviders/%s/%s/" % (gBaseResourcesSection, of, providerName, section))
    else:
        return S_OK(
            gConfig.getValue("%s/%sProviders/%s/%s/%s" % (gBaseResourcesSection, of, providerName, section, option))
        )


def findGenericCloudCredentials(vo=False, group=False):
    """Get the cloud credentials to use for a specific VO and/or group."""
    if not group and not vo:
        return S_ERROR("Need a group or a VO to determine the Generic cloud credentials")
    if not vo:
        vo = Registry.getVOForGroup(group)
        if not vo:
            return S_ERROR("Group %s does not have a VO associated" % group)
    opsHelper = Operations.Operations(vo=vo)
    cloudGroup = opsHelper.getValue("Cloud/GenericCloudGroup", "")
    cloudDN = opsHelper.getValue("Cloud/GenericCloudDN", "")
    if not cloudDN:
        cloudUser = opsHelper.getValue("Cloud/GenericCloudUser", "")
        if cloudUser:
            result = Registry.getDNForUsername(cloudUser)
            if result["OK"]:
                cloudDN = result["Value"][0]
            else:
                return S_ERROR("Failed to find suitable CloudDN")
    if cloudDN and cloudGroup:
        gLogger.verbose("Cloud credentials from CS: %s@%s" % (cloudDN, cloudGroup))
        result = gProxyManager.userHasProxy(cloudDN, cloudGroup, 86400)
        if not result["OK"]:
            return result
        return S_OK((cloudDN, cloudGroup))
    return S_ERROR("Cloud credentials not found")


def getVMTypes(siteList=None, ceList=None, vmTypeList=None, vo=None):
    """Get CE/vmType options filtered by the provided parameters."""

    result = gConfig.getSections("/Resources/Sites")
    if not result["OK"]:
        return result

    resultDict = {}

    grids = result["Value"]
    for grid in grids:
        result = gConfig.getSections("/Resources/Sites/%s" % grid)
        if not result["OK"]:
            continue
        sites = result["Value"]
        for site in sites:
            if siteList is not None and site not in siteList:
                continue
            if vo:
                voList = gConfig.getValue("/Resources/Sites/%s/%s/VO" % (grid, site), [])
                if voList and vo not in voList:
                    continue
            result = gConfig.getSections("/Resources/Sites/%s/%s/Cloud" % (grid, site))
            if not result["OK"]:
                continue
            ces = result["Value"]
            for ce in ces:
                if ceList is not None and ce not in ceList:
                    continue
                if vo:
                    voList = gConfig.getValue("/Resources/Sites/%s/%s/Cloud/%s/VO" % (grid, site, ce), [])
                    if voList and vo not in voList:
                        continue
                result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/Cloud/%s" % (grid, site, ce))
                if not result["OK"]:
                    continue
                ceOptionsDict = result["Value"]
                result = gConfig.getSections("/Resources/Sites/%s/%s/Cloud/%s/VMTypes" % (grid, site, ce))
                if not result["OK"]:
                    result = gConfig.getSections("/Resources/Sites/%s/%s/Cloud/%s/Images" % (grid, site, ce))
                    if not result["OK"]:
                        return result
                vmTypes = result["Value"]
                for vmType in vmTypes:
                    if vmTypeList is not None and vmType not in vmTypeList:
                        continue
                    if vo:
                        voList = gConfig.getValue(
                            "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s/VO" % (grid, site, ce, vmType), []
                        )
                        if not voList:
                            voList = gConfig.getValue(
                                "/Resources/Sites/%s/%s/Cloud/%s/Images/%s/VO" % (grid, site, ce, vmType), []
                            )
                        if voList and vo not in voList:
                            continue
                    resultDict.setdefault(site, {})
                    resultDict[site].setdefault(ce, ceOptionsDict)
                    resultDict[site][ce].setdefault("VMTypes", {})
                    result = gConfig.getOptionsDict(
                        "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s" % (grid, site, ce, vmType)
                    )
                    if not result["OK"]:
                        result = gConfig.getOptionsDict(
                            "/Resources/Sites/%s/%s/Cloud/%s/Images/%s" % (grid, site, ce, vmType)
                        )
                        if not result["OK"]:
                            continue
                    vmTypeOptionsDict = result["Value"]
                    resultDict[site][ce]["VMTypes"][vmType] = vmTypeOptionsDict

    return S_OK(resultDict)


def getVMTypeConfig(site, ce="", vmtype=""):
    """Get the VM image type parameters of the specified queue"""
    tags = []
    reqtags = []
    grid = site.split(".")[0]
    if not ce:
        result = gConfig.getSections("/Resources/Sites/%s/%s/Cloud" % (grid, site))
        if not result["OK"]:
            return result
        ceList = result["Value"]
        if len(ceList) == 1:
            ce = ceList[0]
        else:
            return S_ERROR("No cloud endpoint specified")

    result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/Cloud/%s" % (grid, site, ce))
    if not result["OK"]:
        return result
    resultDict = result["Value"]
    ceTags = resultDict.get("Tag")
    if ceTags:
        tags = fromChar(ceTags)
    ceTags = resultDict.get("RequiredTag")
    if ceTags:
        reqtags = fromChar(ceTags)
    resultDict["CEName"] = ce

    if vmtype:
        result = gConfig.getOptionsDict("/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s" % (grid, site, ce, vmtype))
        if not result["OK"]:
            return result
        resultDict.update(result["Value"])
        queueTags = resultDict.get("Tag")
        if queueTags:
            queueTags = fromChar(queueTags)
            tags = list(set(tags + queueTags))
        queueTags = resultDict.get("RequiredTag")
        if queueTags:
            queueTags = fromChar(queueTags)
            reqtags = list(set(reqtags + queueTags))

    if tags:
        resultDict["Tag"] = tags
    if reqtags:
        resultDict["RequiredTag"] = reqtags
    resultDict["VMType"] = vmtype
    resultDict["Site"] = site
    return S_OK(resultDict)


def getPilotBootstrapParameters(vo="", runningPod=""):
    """Get all of the settings required to bootstrap a cloud instance."""
    op = Operations.Operations(vo=vo)
    result = op.getOptionsDict("Cloud")
    opParameters = {}
    if result["OK"]:
        opParameters = result["Value"]
    opParameters["VO"] = vo
    # FIXME: The majority of these settings can be removed once the old vm-pilot
    #        scripts have been removed.
    opParameters["ReleaseProject"] = op.getValue("Cloud/ReleaseProject", "DIRAC")
    opParameters["ReleaseVersion"] = op.getValue("Cloud/ReleaseVersion", op.getValue("Pilot/Version"))
    opParameters["Setup"] = gConfig.getValue("/DIRAC/Setup", "unknown")
    opParameters["SubmitPool"] = op.getValue("Cloud/SubmitPool")
    opParameters["CloudPilotCert"] = op.getValue("Cloud/CloudPilotCert")
    opParameters["CloudPilotKey"] = op.getValue("Cloud/CloudPilotKey")
    opParameters["pilotFileServer"] = op.getValue("Pilot/pilotFileServer")
    result = op.getOptionsDict("Cloud/%s" % runningPod)
    if result["OK"]:
        opParameters.update(result["Value"])

    # Get standard pilot version now
    if "Version" in opParameters:
        gLogger.warn(
            "Cloud bootstrap version now uses standard Pilot/Version setting. "
            "Please remove all obsolete (Cloud/Version) setting(s)."
        )
    pilotVersions = op.getValue("Pilot/Version")
    if isinstance(pilotVersions, six.string_types):
        pilotVersions = [pilotVersions]
    if not pilotVersions:
        return S_ERROR("Failed to get pilot version.")
    opParameters["Version"] = pilotVersions[0].strip()

    return S_OK(opParameters)
